import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

##############################################################################
# Calculator
##############################################################################
from ase.calculators.emt import EMT

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator

calc = EMT()

##############################################################################
# General settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template, symbols="Au5Ni", confinement_cell=confinement_cell, confinement_corner=confinement_corner
)

# Database
db_path = "db1.db"
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=False,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 400},
    order=2,
)

##############################################################################
# Let get the show running!
##############################################################################


from agox.observer import Observer


class BasicObserver(Observer):
    name = "BasicObserver"

    def __init__(self, order=3):
        Observer.__init__(self, order=order)
        self.add_observer_method(self.basic_observer_method, order=self.order[0], sets={}, gets={})

    @Observer.observer_method
    def basic_observer_method(self, state):
        self.writer(f"Iteration: {self.get_iteration_counter()}: I AM A MODULE.")


basic_observer = BasicObserver()

agox = AGOX(random_generator, database, basic_observer, evaluator)

agox.run(N_iterations=1)



from typing import Optional

import numpy as np


class RateTracker:
    def __init__(self, sample_size: int) -> None:
        self.sample_size = sample_size
        self.acceptance = {i: list() for i in range(sample_size)}
        self.swap_up = {i: list() for i in range(sample_size)}
        self.swap_down = {i: list() for i in range(sample_size)}
        self.rattle_amplitudes = [0] * sample_size

    def update_acceptance(self, sample_index: int, value: int) -> None:
        self.acceptance[sample_index].append(value)

    def update_swap_up(self, sample_index: int, value: int) -> None:
        if sample_index < self.sample_size - 1:
            self.swap_up[sample_index].append(value)

    def update_swap_down(self, sample_index: int, value: int) -> None:
        if sample_index < self.sample_size:
            self.swap_down[sample_index].append(value)

    def get_acceptance_rate(self, sample_index: int, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.acceptance[sample_index][index_slice])

    def get_swap_up_rate(self, sample_index: int, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.swap_up[sample_index][index_slice])

    def get_swap_down_rate(self, sample_index: int, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None) -> float:
        index_slice = slice(start, stop, step)
        return np.mean(self.swap_down[sample_index][index_slice])

    def get_swap_up(self, sample_index: int) -> bool:
        if len(self.swap_up[sample_index]) == 0:
            return False
        return bool(self.swap_up[sample_index][-1])

    def get_swap_down(self, sample_index: int) -> bool:
        if len(self.swap_down[sample_index]) == 0:
            return False
        return bool(self.swap_down[sample_index][-1])

    def get_acceptance(self, sample_index: int) -> bool:
        if len(self.acceptance[sample_index]) == 0:
            return False
        return bool(self.acceptance[sample_index][-1])




from typing import List, Optional, Union

import numpy as np


class Confinement:
    def __init__(
        self,
        cell: np.ndarray = None,
        corner: np.ndarray = None,
        indices: np.ndarray = None,
        pbc: Optional[Union[bool, List[bool]]] = None,
    ) -> None:
        self.cell = np.array(cell)
        self.corner = np.array(corner)

        if indices is None:
            indices = np.array([])
        self.indices = np.array(indices).flatten()

        self.periodicity_setup(pbc)
        self.confined = self.cell is not None and self.corner is not None

    def periodicity_setup(self, pbc: Optional[Union[bool, List[bool]]]) -> None:
        if pbc is None:
            pbc = False
        if isinstance(pbc, bool):
            self.pbc = [pbc] * 3
        elif len(pbc) == 3:
            self.pbc = list(pbc)
        else:
            print("pbc should be list or bool! Setting pbc=False.")
            self.pbc = [False] * 3

        self.hard_boundaries = [not p for p in self.pbc]

        if np.any(self.pbc):
            periodic_cell_vectors = self.cell[:, self.pbc]
            non_periodic_cell_vectors = self.cell[:, self.hard_boundaries]
            if np.any(np.abs(np.matmul(periodic_cell_vectors.T, non_periodic_cell_vectors)) > 0):
                print("---- BOX CONSTRAINT ----")
                print("Periodicity does not work for non-square non-periodic directions!")
                print("------------------------")

        if np.all(self.cell[:, 2] == 0):
            self.dimensionality = 2
            self.effective_cell = self.cell[0:2, 0:2]
            if np.all(self.cell[:, 1] == 0):
                self.effective_cell = self.cell[0:1, 0:1]
                self.dimensionality = 1
        else:
            self.effective_cell = self.cell
            self.dimensionality = 3

    def get_projection_coefficients(self, positions: np.ndarray) -> np.ndarray:
        positions = positions.reshape(-1, 3)
        return np.linalg.solve(
            self.effective_cell.T, (positions - self.corner)[:, 0 : self.dimensionality].T
        ).T.reshape(-1, self.dimensionality)

    def check_confinement(self, positions: np.ndarray) -> np.ndarray:
        """
        Finds the fractional coordinates of the atomic positions in terms of the box defined by the constraint.
        """
        if self.confined:
            C = self.get_projection_coefficients(positions)
            inside = np.all((C > 0) * (C < 1), axis=1)
            return inside
        else:
            return np.ones(positions.shape[0]).astype(bool)

    def get_confinement_limits(self) -> List[np.ndarray]:
        """
        This returns the confinement-limit lists which always assumes a square box.
        """
        conf = [self.corner, self.cell @ np.array([1, 1, 1]) + self.corner]
        return conf

    def _get_box_vector(self) -> np.ndarray:
        return self.cell.T @ np.random.rand(3) + self.corner

    def set_cell(self, cell: np.ndarray, corner: np.ndarray) -> None:
        self.cell = cell
        self.corner = corner
        self.confined = True
        self.confined = self.cell is not None and self.corner is not None

    def set_dimensionality(self, dimensionality: int) -> None:
        self.dimensionality = dimensionality

    def copy(self) -> "Confinement":
        return Confinement(self.cell, self.corner, self.indices, self.pbc)


import functools
from copy import copy
from typing import Callable, Dict, List, Optional

from agox.candidates import StandardCandidate
from agox.module import Module
from agox.observer.observer_handler import ObserverHandler
from agox.observer.observer_method import ObserverMethod
from agox.writer import Writer


class Observer(Module):
    def __init__(
        self,
        gets: List[Dict] = None,
        sets: List[Dict] = None,
        order: List[int] = None,
        **kwargs,
    ) -> None:
        """
        Base-class for classes that act as observers.

        Parameters
        -----------
        gets: dict
        Dict where the keys will be set as the name of attributes with the
        value being the value of the attribute. Used to get something from the
        iteration_cache during a run.

        sets: dict
        Dict where the keys will be set as the name of attributes with the
        value being the value of the attribute. Used to set something from the
        iteration_cache during a run.

        order: int/float
        Specifies the (relative) order of when the observer will be executed,
        lower numbers are executed first.

        sur_name: str
            An additional name added to classes name, can be used to distinguish
            between instances of the same class.
        """
        Module.__init__(self, **kwargs)

        if gets is None:
            gets = dict()
        if isinstance(gets, dict):
            gets = [gets]

        if sets is None:
            sets = dict()
        if isinstance(sets, dict):
            sets = [sets]

        if order is None:
            order = [0]
        if isinstance(order, int) or isinstance(order, float):
            order = [order]

        combined = dict()
        for tuple_of_dicts in [gets, sets]:
            for dict_ in tuple_of_dicts:
                for key, item in dict_.items():
                    combined[key] = item
        for key, value in combined.items():
            if key not in self.__dict__.keys():
                self.__dict__[key] = value
            else:
                raise NameError("Key {} already exists in the class.".format(key))

        self.set_keys = sum([list(set_dict.keys()) for set_dict in sets], [])
        self.set_values = sum([list(set_dict.values()) for set_dict in sets], [])
        self.get_keys = sum([list(get_dict.keys()) for get_dict in gets], [])
        self.get_values = sum([list(get_dict.values()) for get_dict in gets], [])
        self.gets = gets
        self.sets = sets

        self.order = order

        self.observer_methods = {}
        self.observer_handler_identifiers = []

        self.iteration_counter = None
        self.candidate_instanstiator = StandardCandidate

    def add_observer_method(
        self, method: Callable, sets: Dict, gets: Dict, order: int, handler_identifier: str = "any"
    ) -> None:
        """
        Adds an observer method that will later be attached with 'self.attach'.

        Parameters
        -----------
        method: method
            The function/method that will be called by the observer-loop.
        sets: dict
            Dict containing the keys that the method will set in the cache.
        gets: dict
            Dict containing the keys that the method will get from the cache.
        """
        observer_method = ObserverMethod(self.__name__, method.__name__, method, gets, sets, order, handler_identifier)
        self.observer_methods[observer_method.key] = observer_method

    def remove_observer_method(self, observer_method: ObserverMethod) -> None:
        """
        Remove an observer_method from the internal dict of methods, if this is
        done before calling 'attach' it will not be added as an observer.

        Parameters
        -----------
        observer_method: object
            An instance of ObserverMethod.
        """
        key = observer_method.key
        if key in self.observer_methods.keys():
            del self.observer_methods[key]

    def update_order(self, observer_method: ObserverMethod, order: int) -> None:
        """
        Change the order of a method. Not really tested in practice.

        Parameters
        -----------
        observer_method: object
            An instance of ObserverMethod.

        order: float
            New order to set.
        """
        key = observer_method.key
        assert key in self.observer_methods.keys()
        self.observer_methods[key].order = order

    def attach(self, handler: ObserverHandler) -> None:
        """
        Attaches all ObserverMethod's as to observer-loop of the ObserverHandler
        object 'main'

        Parameters
        ----------
        handler: object
            An instance of an ObserverHandler.
        """
        for observer_method in self.observer_methods.values():
            if (
                observer_method.handler_identifier == "any"
                or observer_method.handler_identifier == handler.handler_identifier
            ):
                observer_method.set_handler(handler.name)
                handler.attach_observer(observer_method)
                self.observer_handler_identifiers.append(handler.handler_identifier)

    def reset_observer(self) -> None:
        """
        Reset the observer_methods dict, removing all observer methods.
        """
        self.observer_methods = {}

    ############################################################################
    # State
    ############################################################################

    def observer_method(func) -> Callable:  # noqa
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):  # noqa
            self.writer.write_header(self.name)
            state = kwargs.get("state")
            self.state_update(state)
            results = func(self, *args, **kwargs)
            return results

        return wrapper

    def state_update(self, state) -> None:  # noqa
        self.set_iteration_counter(state.get_iteration_counter())

    ############################################################################
    # Iteration counter methods:
    ############################################################################

    def get_iteration_counter(self) -> int:
        return self.iteration_counter

    def set_iteration_counter(self, iteration_count: int) -> None:
        self.iteration_counter = iteration_count

    def check_iteration_counter(self, count: int) -> bool:
        if self.iteration_counter is None:
            return True
        return self.iteration_counter >= count

    ############################################################################
    # Checker
    ############################################################################

    def do_check(self, **kwargs) -> bool:
        """
        Check if the method will run or just do nothing this iteration.

        Returns
        -------
        bool
            True if function will run, False otherwise.w
        """
        return True

    ############################################################################
    # Candidate instanstiator
    ############################################################################

    def set_candidate_instanstiator(self, candidate_instanstiator: StandardCandidate) -> None:
        self.candidate_instanstiator = copy(candidate_instanstiator)



from abc import ABC, abstractmethod  # noqa: N999

from agox.utils.constraints import BoxConstraint, ConstraintManager
import numpy as np
from ase.constraints import FixAtoms

from agox.candidates import StandardCandidate
from agox.helpers.confinement import Confinement
from agox.module import Module
from agox.utils.constraints.box_constraint import BoxConstraint



class EnvironmentBaseClass(ABC, Module):
    """
    Base class for all environments.

    Environments are used to define the environment in which the candidates are
    generated. This includes confinement, constraints, and other settings.

    Parameters
    ----------
    confinement_cell : np.ndarray, optional
        The cell of the confinement, by default None.
    confinement_corner : np.ndarray, optional
        The corner of the confinement, by default None.
    constraints : list, optional
        List of constraints, by default [].
    use_box_constraint : bool, optional
        If True, a box constraint is used, by default True.
    box_constraint_pbc : list, optional
        List of booleans, by default [False]*3.
    fix_template : bool, optional
        If True, the template is fixed, by default True.
    """

    name = "Environment"

    def __init__(
        self,
        confinement_cell: np.ndarray = None,
        confinement_corner: np.ndarray =None,
        constraints: list | None = None,
        use_box_constraint: bool = True,
        box_constraint_pbc: list[bool, bool, bool] | None = None,
        fix_template: bool = True,
        surname: str | None = None,
        **kwargs,
    ) -> None:
        Module.__init__(self, surname=surname, **kwargs)

        self.confinement_cell = confinement_cell
        self.confinement_corner = confinement_corner
        self.constraints = constraints if constraints is not None else []
        self.use_box_constraint = use_box_constraint
        self.box_constraint_pbc = box_constraint_pbc if box_constraint_pbc is not None else [False] * 3
        self.fix_template = fix_template
        self.constraints = []

    @abstractmethod
    def get_template(self, **kwargs):  # pragma: no cover
        pass

    @abstractmethod
    def get_numbers(self, numbers, **kwargs):
        pass

    @abstractmethod
    def environment_report(self):
        pass

    def get_missing_indices(self):
        return np.arange(len(self._template), len(self._template) + len(self._numbers))

    def get_confinement_cell(self, distance_to_edge=3):
        if self.confinement_cell is not None:
            confinement_cell = self.confinement_cell
        elif self._template.pbc.all() is False:
            confinement_cell = self._template.get_cell().copy() - np.eye(3) * distance_to_edge * 2
        else:
            # Find the directions that are not periodic:
            confinement_cell = self._template.get_cell().copy() - np.eye(3) * distance_to_edge * 2
            directions = np.argwhere(self._template.pbc == True)
            for d in directions:
                confinement_cell[d, :] = self._template.get_cell()[d, :]

        return confinement_cell

    def get_confinement_corner(self, distance_to_edge=3):
        if self.confinement_cell is not None:
            confinement_corner = self.confinement_corner
        elif self._template.pbc.all() is False:
            confinement_corner = np.ones(3) * distance_to_edge
        else:
            # Find the directions that are not periodic:
            confinement_corner = np.ones(3) * distance_to_edge
            directions = np.argwhere(self._template.pbc == True)
            for d in directions:
                confinement_corner[d] = 0
        return confinement_corner

    def get_confinement(self, as_dict: bool = True):
        confinement = Confinement(self.confinement_cell, self.confinement_corner)
        if as_dict:
            return {'confinement': confinement}
        return confinement

    def get_box_constraint(self):
        confinement_cell = self.get_confinement_cell()
        confinement_corner = self.get_confinement_corner()
        return BoxConstraint(
            confinement_cell=confinement_cell,
            confinement_corner=confinement_corner,
            indices=self.get_missing_indices(),
            pbc=self.box_constraint_pbc,
        )

    def get_constraints(self):
        manager = ConstraintManager()
        if self.use_box_constraint:
            manager.add_constraint(self.get_box_constraint(), mode="optimize")
        if self.fix_template:
            manager.add_constraint(FixAtoms(indices=[]), mode="template")

        for constraint, mode in self.constraints:
            manager.add_constraint(constraint, mode=mode)
        return manager

    def add_constraint(self, constraint, mode=None):
        self.constraints.append((constraint, mode))

    def convert_to_candidate_object(self, template):
        template = StandardCandidate(
            template=template,
            positions=template.positions,
            numbers=template.numbers,
            cell=template.cell,
            info=template.info,
        )
        return template

    def plot(self, name="environment_plot.png"):
        import matplotlib.pyplot as plt

        from agox.utils.matplotlib_utils import use_agox_mpl_backend
        from agox.utils.plot import plot_atoms, plot_cell

        use_agox_mpl_backend()

        atoms = self.get_template()
        atoms.set_constraint(self.get_constraints())

        fig, axs = plt.subplots(1, 3, figsize=(15, 5))

        for ax, plane in zip(axs, ["xy", "xz", "yz"]):
            plot_atoms(ax, atoms, plane=plane, plot_constraint=True, repeat=True)
            plot_cell(ax, atoms.cell, plane=plane, collection_kwargs=dict(edgecolors="black", linestyles="dotted"))
            plot_cell(
                ax,
                self.confinement_cell,
                plane=plane,
                offset=self.confinement_corner,
                collection_kwargs=dict(edgecolors="red", linestyles="dashed"),
            )

            # Equal aspect ratio
            ax.set_aspect("equal", "box")

        fig.savefig(name, bbox_inches="tight")
        plt.close()



import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error, median_absolute_error, r2_score


def get_metrics(true, pred):
    """Get metrics for a regression problem.

    Args:
        true (np.ndarray): true values
        pred (np.ndarray): predicted values

    Returns:
        dict: dictionary of metrics
    """

    metrics = {}
    metrics["mae"] = mean_absolute_error(true, pred)
    metrics["rmse"] = np.sqrt(mean_squared_error(true, pred))
    metrics["r2"] = r2_score(true, pred)
    metrics["mdae"] = median_absolute_error(true, pred)
    return metrics



import numpy as np
from ase.constraints import FixedLine, FixedPlane
from ase.md.langevin import Langevin
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary, ZeroRotation
from ase.units import fs

from agox.generators.ABC_generator import GeneratorBaseClass


class MDgenerator(GeneratorBaseClass):
    """
    Does a molecular dynamics simulation on the candidate structure to propose a new candidate.

    Parameters:
    -----------
    calculator : ASE calculator
        Calculator for the MD simulation.
    thermostat : ASE thermostat
        MD program used.
    thermostat_kwargs : dict
        Settings for MD program.
    start_settings : list
        List of start settings for the MD simulation.
    start_settings_kwargs : list
        List of dictionaries with settings for start settings.
    temperature_program : list
        List of tuples with temperature and steps for MD program if temperature is modifiable during simulation.
    set_start_settings : bool
        Whether to set start settings or not.
    constraints : list
        Constraints besides fixed template and 1D/2D constraints.
    check_template : bool
        Check if template atoms moved during MD simulation.
    """

    name = "MDgenerator"

    def __init__(
        self,
        calculator,
        thermostat=Langevin,
        thermostat_kwargs={"timestep": 1.0 * fs, "temperature_K": 10, "friction": 0.05},
        start_settings=[MaxwellBoltzmannDistribution, ZeroRotation, Stationary],
        start_settings_kwargs=[{"temperature_K": 100}, {}, {}],
        temperature_program=[(500, 10), (100, 10)],
        set_start_settings=False,
        constraints=[],
        check_template=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.calculator = calculator  # Calculator for MD simulation
        self.thermostat = thermostat  # MD program used
        self.thermostat_kwargs = thermostat_kwargs  # Settings for MD program
        self.set_start_settings = set_start_settings
        self.start_settings = start_settings
        self.start_settings_kwargs = start_settings_kwargs
        self.temperature_program = (
            temperature_program  # (temp in kelvin, steps) for MD program if temperature is modifiable during simulation
        )
        self.constraints = constraints  # Constraints besides fixed template and 1D/2D constraints

        self.check_template = check_template  # Check if template atoms moved during MD simulation

    def _get_candidates(self, candidate, parents, environment):
        candidate.set_calculator(self.calculator)

        self.remove_constraints(
            candidate
        )  # All constraints are removed from candidate before applying self.constraints to ensure only constraints set by user are present during MD simulation
        self.apply_constraints(candidate)
        self.molecular_dynamics(candidate)
        self.remove_constraints(
            candidate
        )  # All constraints are removed after MD simulation to not interfere with other AGOX modules

        candidate.add_meta_information("description", self.name)

        return [candidate]

    def molecular_dynamics(self, candidate):
        """Runs the molecular dynamics simulation and applies/removes constraints accordingly"""

        if self.set_start_settings:
            for setting, kwargs in zip(self.start_settings, self.start_settings_kwargs):
                setting(candidate, **kwargs)

        temperature_program = []
        for temp, steps in self.temperature_program:
            t = np.random.normal(temp, 0.1 * temp)
            temperature_program.append((t, steps))

        dyn = self.thermostat(candidate, **self.thermostat_kwargs)

        if self.check_template:
            positions_before = candidate.template.positions

        for temp, steps in temperature_program:
            dyn.set_temperature(temperature_K=temp)
            dyn.run(steps)

        if self.check_template:
            positions_after = candidate.template.positions

            if np.array_equal(positions_before, positions_after):
                self.writer("Template positions were not altered by MD simulation")
                # print('Template positions were not altered by MD simulation')
            else:
                self.writer("Template positions were altered by MD simulation")
                # print('Template positions were altered by MD simulation')

    def apply_constraints(self, candidate):
        """Applies constraints manually set and based on dimensionality of confinement cell"""

        constraints = self.constraints  # Add any passed constraints immediately
        dimensionality = self.get_dimensionality()
        if dimensionality == 1 or dimensionality == 2:  # Ensures movement within 1D or 2D confinement cell
            constraints.append(self.get_dimensionality_constraints(candidate))

        candidate.set_constraint(constraints)

    def get_dimensionality_constraints(self, candidate):
        """Depending on the dimensionality this either sets fixed line or fixed plane.
        Similar to how rattle needs dimensionality specified to match with confinement cell
        """

        template = candidate.get_template()
        n_template = len(template)
        n_total = len(candidate)
        dimensionality = self.get_dimensionality()
        if dimensionality == 1:
            return FixedLine(
                indices=np.arange(n_template, n_total), direction=[1, 0, 0]
            )  # Generator assumes 1D search happens in X direction
        if dimensionality == 2:
            return FixedPlane(
                indices=np.arange(n_template, n_total), direction=[0, 0, 1]
            )  # Generator assumes 2D search happens in XY-plane

    def remove_constraints(self, candidate):
        """Removes all constraints from candidate"""

        candidate.set_constraint([])

    def get_number_of_parents(self, sampler):
        return 1



import numpy as np
from agox.utils.ray_utils import RayPoolUser
from agox.module import Module
import torch


def net_forward(parallel_class, x):
    return parallel_class.net(torch.tensor(x, dtype=torch.float64)).detach().numpy() + parallel_class.bias


class SimpleNet(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = torch.nn.Linear(100, 100, dtype=torch.float64)
        self.fc2 = torch.nn.Linear(100, 1, dtype=torch.float64)

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class ParallelClass(RayPoolUser):
    dynamic_attributes = ["net"]

    def __init__(self, net, bias, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.net = net
        self.bias = bias

        self.ray_key = self.pool_add_module(self)

    def do_parallel(self, X):
        # Make the jobs:
        N_jobs = X.shape[0]
        modules = [[self.ray_key]] * N_jobs
        args = [[x] for x in X]
        kwargs = [{} for _ in range(N_jobs)]

        # Run the jobs:
        results = self.pool_map(net_forward, modules, args, kwargs)

        return np.array(results)

    def do_serial(self, X):
        return self.net(torch.tensor(X, dtype=torch.float64)).detach().numpy() + self.bias


# Make an instance of the class:
net = SimpleNet()
X = np.random.rand(10, 100)
parallel_class = ParallelClass(net, bias=1, cpu_count=4)

# Predictions:
results_parallel = parallel_class.do_parallel(X)
results_serial = parallel_class.do_serial(X)

# Compare to serial:
np.testing.assert_allclose(results_parallel, results_serial)  # Suceeds.

# Scramblee network:
net = SimpleNet()
parallel_class.net = net

# Synchronize:
parallel_class.pool_synchronize(writer=print)

# Predictions:
results_parallel = parallel_class.do_parallel(X)
results_serial = parallel_class.do_serial(X)

# Compare to serial:
np.testing.assert_allclose(results_parallel, results_serial)  # Succeeds

parallel_class.bias += 1

# Synchronize:
parallel_class.pool_synchronize(writer=print)

# Predictions:
results_parallel = parallel_class.do_parallel(X)
results_serial = parallel_class.do_serial(X)

# Compare to serial:
np.testing.assert_allclose(results_parallel, results_serial)  # Fails



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from agox.analysis.criterion.base_criterion.discrete_distribution import DiscreteDistribution
from agox.analysis.search_data import SearchData


class BaseCriterion(ABC):
    @abstractmethod
    def compute(self, SearchData) -> Tuple[np.array, np.array]:
        """
        Compute the times and events for the boolean metric.
        """
        pass

    def __call__(self, search_data: SearchData) -> DiscreteDistribution:
        times, events = self.compute(search_data)
        return DiscreteDistribution(times, events)



"""
Collectors are used to collect multiple candidates generated by generators.
This is a task that can be parallelized over the number of generated candidates,
and this is done using parallelization from Ray.
"""
# ruff: noqa: I001, E402

from typing import TypeAlias
from .ABC_collector import CollectorBaseClass
Collector: TypeAlias = CollectorBaseClass

from agox.collectors.replica_exchange import ReplicaExchangeCollector
from agox.collectors.ray_collector import ParallelCollector
from agox.collectors.standard import StandardCollector

__all__ = [
        "CollectorBaseClass",
        "Collector",
        "StandardCollector", 
        "ParallelCollector", 
        "ReplicaExchangeCollector",
        ]



import numpy as np
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.optimize import BFGS

from agox.postprocessors.ABC_postprocess import PostprocessBaseClass


class RelaxPostprocess(PostprocessBaseClass):
    """
    Relaxes candidates using a local optimization algorithm.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    fix_template : bool
        If True, the template atoms are fixed during relaxation.
    constraints : list
        List of constraints to apply during relaxation.
    start_relax : int
        The iteration to start relaxing candidates.
    """

    name = "PostprocessRelax"

    def __init__(
        self,
        model=None,
        optimizer=None,
        optimizer_kwargs={"logfile": None},
        optimizer_run_kwargs={"fmax": 0.05, "steps": 200},
        constraints=[],
        start_relax=1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.optimizer = BFGS if optimizer is None else optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimizer_run_kwargs = optimizer_run_kwargs
        self.start_relax = start_relax
        self.model = model

        # Constraints:
        self.constraints = constraints
    
    def postprocess(self, candidate):
        initial_candidate = candidate.copy()
        candidate.set_calculator(self.model)
        self.apply_constraints(candidate)
        optimizer = self.optimizer(candidate, **self.optimizer_kwargs)

        try:
            optimizer.run(**self.optimizer_run_kwargs)
        except Exception as e:
            print("Relaxation failed with exception: {}".format(e))
            return initial_candidate

        candidate.add_meta_information("relaxation_steps", optimizer.get_number_of_steps())

        results = {prop: val for prop, val in candidate.calc.results.items() if prop in all_properties}
        candidate.calc = SinglePointCalculator(candidate, **results)

        print(f"Relaxed for {optimizer.get_number_of_steps()} steps")

        self.remove_constraints(candidate)
        return candidate

    def do_check(self, **kwargs):
        return self.check_iteration_counter(self.start_relax) * self.model.ready_state

    ####################################################################################################################
    # Constraints
    ####################################################################################################################

    def apply_constraints(self, candidate):
        candidate.set_constraint(self.constraints)

    def remove_constraints(self, candidate):
        candidate.set_constraint([])


import numpy as np

from agox.databases.database_concurrent import ConcurrentDatabase


class OrderedDatabase(ConcurrentDatabase):
    def restore_to_memory(self):
        strucs = self.get_all_structures_data()
        if self.store_meta_information:
            all_meta_info = self.read_all_key_value_pairs()
        else:
            all_meta_info = {}

        all_candidates = []
        for struc in strucs:
            all_candidates.append(self.db_to_candidate(struc, meta_dict=all_meta_info.get(struc["id"], None)))

        iteration = self.get_iteration_counter()

        new_candidates = [
            c for c in all_candidates if c.get_meta_information("iteration") > iteration - self.sync_frequency
        ]

        old_candidates = [
            c for c in self.candidates if c.get_meta_information("iteration") <= iteration - self.sync_frequency
        ]

        worker_numbers = np.array(
            [c.get_meta_information("worker_number") != self.worker_number for c in new_candidates], dtype=int
        )
        iterations = np.array([c.get_meta_information("iteration") for c in new_candidates])
        relax_indices = np.array([c.get_meta_information("relax_index") for c in new_candidates])
        indices = np.lexsort((iterations, worker_numbers))

        self.candidates = old_candidates + [new_candidates[i] for i in indices]


        for c in self.candidates:
            self.writer(
                f"worker number: {c.get_meta_information('worker_number')}, iteration: {c.get_meta_information('iteration')}, relax_index: {c.get_meta_information('relax_index')}"
            )



from typing import Literal

import numpy as np

from agox.analysis import SearchData
from agox.models.descriptors import DescriptorBaseClass

from .property import ListPropertyData, Property


class DescriptorProperty(Property):

    def __init__(self, descriptor: DescriptorBaseClass, **kwargs) -> None:
        """
        Parameters
        ----------
        descriptor : DescriptorBaseClass
            The descriptor to compute.
        """
        super().__init__(**kwargs)
        self.descriptor = descriptor

    def compute(self, search_data: SearchData) -> ListPropertyData:
        """ """
        restarts = search_data.get_restarts()
        descriptor_list = []

        for restart in restarts:
            features = self.descriptor.get_features(restart.candidates)
            descriptor_list.append(features)

        axis_name, indices = self.get_time_axis(search_data)

        identifiers = search_data.get_all_identifiers()

        return ListPropertyData(
            name=f"Descriptor-{self.descriptor.name}",
            data=descriptor_list,
            shape=("Restarts", axis_name),
            list_axis=(identifiers, indices),
        )



import os
import os.path
import pickle
import sqlite3
from collections import OrderedDict
from time import sleep
from typing import Dict, List, Optional

import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
import time

# Should match init_statements!!
from agox.candidates import Candidate, StandardCandidate
from agox.databases.database_utilities import blob, deblob, nothing, now

from .ABC_database import DatabaseBaseClass


class Database(DatabaseBaseClass):
    """Database module"""

    init_statements = [
        """create table structures (
    id integer primary key autoincrement,
    ctime real,
    positions blob,
    energy real,
    type blob,
    cell blob,
    forces blob, 
    pbc blob,
    template_indices blob,
    iteration int,
    cache_key text
    )""",
        """CREATE TABLE text_key_values (
    key TEXT,
    value TEXT,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",
        """CREATE TABLE float_key_values (
    key TEXT,
    value REAL,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",
        """CREATE TABLE int_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",
        """CREATE TABLE boolean_key_values (
    key TEXT,
    value INTEGER,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",
        """CREATE TABLE other_key_values (
    key TEXT,
    value BLOB,
    id INTEGER,
    FOREIGN KEY (id) REFERENCES systems(id))""",
    ]

    # Pack: Positions, energy, type, cell, forces, pbc, template_indices, iteration
    # Unpack: ID, time, -//-
    pack_functions = [blob, nothing, blob, blob, blob, blob, blob, nothing, nothing]
    unpack_functions = [nothing, nothing, deblob, nothing, deblob, deblob, deblob, deblob, deblob, nothing, nothing]

    name = "Database"

    def __init__(
        self,
        filename: str = "db.db",
        initialize: bool = False,
        write_interval: int = 1,
        call_initialize: bool = True,
        store_meta_information: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        # File-based stuff:
        self.filename = filename
        if initialize and os.path.exists(filename):
            os.remove(filename)
        self.con = sqlite3.connect(filename, timeout=600)
        self.write_interval = write_interval

        # Important that this matches the init_statements list.
        self.storage_keys = [
            "positions",
            "energy",
            "type",
            "cell",
            "forces",
            "pbc",
            "template_indices",
            "iteration",
            "cache_key",
        ]

        # Memory-based stuff:
        self.candidate_instanstiator = StandardCandidate
        self.candidate_energies = []

        # Whether or not to save and retrieve meta information from the database.
        self.store_meta_information = store_meta_information
        self.meta_dict_list = []

        if call_initialize:
            self._initialize()

        self.start_time = time.time()

    ####################################################################################################################
    # Memory-based methods:
    ####################################################################################################################

    def store_candidate(self, candidate: Candidate, accepted: bool = True, write: bool = True) -> None:
        candidate = self._copy_candidate(candidate)

        if accepted:
            self.candidates.append(candidate)
            self.candidate_energies.append(candidate.get_potential_energy())
        if write:
            self.write(candidate)

    def get_best_energy(self) -> float:
        return np.min(self.candidate_energies)

    def add_time_stamp(self, candidate: Candidate) -> None:
        if "time_stamp" not in candidate.meta_information.keys():
            time_stamp = time.time() - self.start_time
            candidate.add_meta_information("time_stamp", time_stamp)

    ####################################################################################################################
    # File-based methods:
    ####################################################################################################################

    def _init_storage(self) -> None:
        self.positions = []
        self.atom_numbers = []
        self.energies = []
        self.cells = []
        self.forces = []
        self.pbc = []
        self.template_indices = []

        self.storage_dict = OrderedDict()
        for key in self.storage_keys:
            self.storage_dict[key] = []

        self.number_of_rows = 0
        self.meta_dict_list = []

    def _initialize(self) -> None:
        # Check if structures are in the database, otherwise initialize tables
        cur = self.con.execute('select count(*) from sqlite_master where name="structures"')
        try:
            if cur.fetchone()[0] == 0:
                for statement in self.init_statements:
                    self.con.execute(statement)
                self.con.commit()
        except sqlite3.OperationalError:
            pass
        self._init_storage()
        self.con.row_factory = sqlite3.Row

    def store_information(self, candidate: Candidate) -> None:
        if candidate is not None:
            candidate.print_properties(self.writer, self.get_iteration_counter())
            e = candidate.calc.results.get("energy", 0)
            f = candidate.calc.results.get("forces", np.zeros(candidate.positions.shape))

            self.storage_dict["energy"].append(e)
            self.storage_dict["forces"].append(f)
            self.storage_dict["positions"].append(np.array(candidate.positions, dtype=np.float64))
            self.storage_dict["type"].append(np.array(candidate.numbers, dtype=np.float64))
            self.storage_dict["cell"].append(np.array(candidate.cell, dtype=np.float64))
            self.storage_dict["pbc"].append(np.array(candidate.pbc.astype(int), dtype=np.float64))
            self.storage_dict["template_indices"].append(np.array(candidate.get_template_indices(), dtype=np.float64))
            self.storage_dict["iteration"].append(candidate.meta_information.get("iteration", self.get_iteration_counter()))
            self.storage_dict["cache_key"].append(candidate.cache_key)
            self.number_of_rows += 1

            if self.store_meta_information:
                self.meta_dict_list.append(candidate.meta_information)

    def get_row_to_write(self, index: int) -> tuple:
        row = [now()]
        for (key, value), func in zip(self.storage_dict.items(), self.pack_functions):
            row.append(func(value[index]))

        return tuple(row)

    def get_last_id(self, cur: sqlite3.Cursor) -> int:
        cur.execute('SELECT seq FROM sqlite_sequence WHERE name="structures"')
        result = cur.fetchone()
        if result is not None:
            iden = result[0]
            return iden
        else:
            return 0

    def write(
        self, candidate: Optional[Candidate] = None, force_write: bool = False, already_stored: bool = False
    ) -> None:
        if not already_stored:
            self.store_information(candidate)

        if self.number_of_rows == self.write_interval or force_write:
            try:
                for row_index in range(self.number_of_rows):
                    cur = self.con.cursor()
                    row = self.get_row_to_write(row_index)
                    q = "NULL," + ", ".join("?" * len(row))

                    cur.execute("insert into structures values ({})".format(q), row)

                    self.last_used_id = self.get_last_id(cur)

                    if self.store_meta_information:
                        cur = self.con.cursor()
                        iden = self.last_used_id
                        self.write_key_value_pairs(cur, self.meta_dict_list[row_index], id=iden)
                    self.con.commit()

            except sqlite3.OperationalError as error:
                print("Encountered error: {}".format(error))
                sleep(1)
                self.write(candidate=candidate, force_write=force_write, already_stored=True)
            self._init_storage()

    def db_to_atoms(self, structure: Dict) -> Atoms:
        """

        Converts a database representation (dictionary) of a structure to an ASE atoms object

        Parameters
        ----------
        structure :  database representation of a structure

        Returns
        -------
        struc : ASE Atoms object

        """
        e = structure["energy"]
        f = structure.get("forces", 0).reshape(-1, 3)
        pos = structure["positions"]
        num = structure["type"]
        cell = structure["cell"]
        pbc = structure.get("pbc", None)

        atoms = Atoms(symbols=num, positions=pos, cell=cell, pbc=pbc)

        calc = SinglePointCalculator(atoms, energy=e, forces=f)
        atoms.calc = calc
        return atoms

    def db_to_candidate(self, structure: Dict, meta_dict: Optional[Dict] = None) -> Candidate:
        e = structure["energy"]
        f = structure.get("forces", 0).reshape(-1, 3)
        pos = structure["positions"]
        num = structure["type"]
        cell = structure["cell"]
        pbc = structure.get("pbc", None)
        template_indices = structure.get("template_indices", None)
        cache_key = structure.get("cache_key", None)

        if hasattr(self, "template") and template_indices is None:
            template = self.template
        else:
            template = None

        candidate = self.candidate_instanstiator(
            symbols=num, positions=pos, cell=cell, pbc=pbc, template=template, template_indices=template_indices
        )
        candidate.cache_key = cache_key
        calc = SinglePointCalculator(candidate, energy=e, forces=f)
        candidate.set_calculator(calc)

        if self.store_meta_information:
            if meta_dict is None:
                restored_dict = self.read_key_value_pairs(id=structure["id"])
                for key, value in restored_dict.items():
                    candidate.add_meta_information(key, value)
            else:
                candidate.meta_information.update(meta_dict)

        candidate.add_meta_information("iteration", structure.get("iteration", 0))
        candidate.add_meta_information("database_id", structure["id"])

        return candidate

    def get_data_from_row(self, row: Dict) -> Dict:
        d = {}
        for key, value, func in zip(row.keys(), row, self.unpack_functions):
            d[key] = func(value)

        d["positions"] = d["positions"].reshape(-1, 3)
        d["cell"] = d["cell"].reshape(3, 3)

        return d

    def get_all_structures_data(self) -> List:
        cursor = self.con.execute("SELECT * from structures")
        structures = []
        for row in cursor.fetchall():
            d = self.get_data_from_row(row)
            structures.append(d)
        return structures

    def get_structure_data(self, i: int) -> Dict:
        t = (str(int(i)),)
        cursor = self.con.execute("SELECT * from structures WHERE id=?", t)
        row = cursor.fetchone()
        d = self.get_data_from_row(row)
        return d

    def get_all_energies(self) -> np.ndarray:
        cursor = self.con.execute("SELECT energy from structures")
        energies = []
        for row in cursor.fetchall():
            energies.append(row["energy"])
        return np.array(energies)

    def get_best_structure(self) -> Atoms:
        structure_data = self.get_all_structures_data()
        energies = [c["energy"] for c in structure_data]
        idx = energies.index(min(energies))
        best_struc = self.db_to_atoms(structure_data[idx])
        return best_struc

    def get_structure(self, index: int) -> Atoms:
        cand = self.get_structure_data(index)
        struc = self.db_to_atoms(cand)

        return struc

    ####################################################################################################################
    # Meta-information related.
    ####################################################################################################################

    def restore_to_memory(self) -> None:
        strucs = self.get_all_structures_data()
        if self.store_meta_information:
            all_meta_info = self.read_all_key_value_pairs()
        else:
            all_meta_info = {}

        candidates = []
        for struc in strucs:
            candidates.append(self.db_to_candidate(struc, meta_dict=all_meta_info.get(struc["id"], None)))
        self.candidates = candidates
        self.candidate_energies = [atoms.get_potential_energy() for atoms in candidates]

    def restore_to_trajectory(self) -> List[Atoms]:
        strucs = self.get_all_structures_data()
        atoms_objects = []
        for structure in strucs:
            atoms_objects.append(self.db_to_atoms(structure))
        return atoms_objects

    ####################################################################################################################
    # Meta-information related.
    ####################################################################################################################

    def write_key_value_pairs(self, cur, dictionary, id=0): # noqa
        text_pairs = []
        float_pairs = []
        int_pairs = []
        bool_pairs = []
        other_pairs = []

        for key, value in dictionary.items():
            if isinstance(value, str):
                text_pairs += [(key, value, id)]
            elif isinstance(value, float):
                float_pairs += [(key, value, id)]
            elif isinstance(value, int):
                int_pairs += [(key, value, id)]
            elif isinstance(value, bool):
                bool_pairs += [(key, value, id)]
            else:
                value_converted = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
                other_pairs += [(key, value_converted, id)]

        cur.executemany("INSERT INTO text_key_values VALUES (?, ?, ?)", text_pairs)
        cur.executemany("INSERT INTO float_key_values VALUES (?, ?, ?)", float_pairs)
        cur.executemany("INSERT INTO int_key_values VALUES (?, ?, ?)", int_pairs)
        cur.executemany("INSERT INTO boolean_key_values VALUES (?, ?, ?)", bool_pairs)
        cur.executemany("INSERT INTO other_key_values VALUES (?, ?, ?)", other_pairs)

    def read_key_value_pairs(self, id: int = 0) -> Dict: # noqa
        dict_recreation = {}
        tables = ["text_key_values", "float_key_values", "int_key_values", "boolean_key_values", "other_key_values"]
        functions = [nothing, nothing, nothing, bool, pickle.loads]

        cur = self.con.cursor()
        for table, func in zip(tables, functions):
            cur.execute(f"SELECT * FROM {table} WHERE id=?", (id,))
            a = cur.fetchall()
            for key, value, _ in a:
                try:
                    converted_value = func(value)
                except Exception as _:
                    converted_value = deblob(value)

                dict_recreation[key] = converted_value

        return dict_recreation

    def read_all_key_value_pairs(self) -> Dict:
        cursor = self.con.cursor()

        tables = ["text_key_values", "float_key_values", "int_key_values", "boolean_key_values", "other_key_values"]
        functions = [nothing, nothing, nothing, bool, pickle.loads]

        dict_of_dicts = {}  # Uses the id as the key to the meta information dict for each candidate. (Id from structures table).

        for table, func in zip(tables, functions):
            cursor.execute(f"SELECT * FROM {table}")
            rows = cursor.fetchall()

            for key, value, iden in rows:
                try:
                    converted_value = func(value)
                except Exception as _:
                    converted_value = deblob(value)

                if iden in dict_of_dicts.keys():
                    dict_of_dicts[iden][key] = converted_value
                else:
                    dict_of_dicts[iden] = {key: converted_value}

        return dict_of_dicts

    def update_meta_information(self, meta_dict: Dict, database_index: int, memory_index: int) -> None:
        """

        Update meta-information of the candidate at database_index 'id' in the structures
        table.

        Parameters
        ----------
        meta_dict : dict
            Dict containing the information to update - if keys are present that
            are already present in the database they are overwritten with the new
            values and if new keys are present they are added to the database.
        index : int
            The index (id) of the structure in the structures table.
        """
        if database_index is not None:
            cur = self.con.cursor()
            self.write_key_value_pairs(cur, meta_dict, id=database_index)
            self.con.commit()
        if memory_index is not None:
            self.candidates[memory_index].meta_information.update(meta_dict)



from agox.analysis.criterion.base_criterion import BaseCriterion
from agox.analysis.criterion.distance import DistanceCriterion
from agox.analysis.criterion.threshold import ThresholdCriterion

__all__ = ["BaseCriterion", "ThresholdCriterion", "DistanceCriterion"]



from agox.utils.thermodynamics.thermodynamics_data import ThermodynamicsData # noqa
from agox.utils.thermodynamics.gibbs import gibbs_free_energy

__all__ = ["ThermodynamicsData", "gibbs_free_energy"]



# Copyright 2021-2024, Mads-Peter V. Christiansen, Bjørk Hammer, Nikolaj Rønne.
# This file is part of AGOX.
# AGOX is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# AGOX is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should
# have received a copy of the GNU General Public License along with AGOX. If not, see <https://www.gnu.org/licenses/>.

from collections import OrderedDict
from copy import deepcopy
from typing import Any, List, Literal

from agox.observer import Observer


class State:
    def __init__(self) -> None:
        """
        State object.

        Attributes
        ----------
        cache: dict
            Data communicated between modules is stored in the cache.
        iteration_counter: int
            Keeps track of the number of iterations.
        convergence: bool
            Convergence status, if True the iteration-loop is halted.
        """

        self.cache = {}
        self.execution_times = {}
        self.iteration_counter = 1
        self.converged = False

    def get_iteration_counter(self) -> int:
        """
        Returns
        -------
        int
            The current iteration number.
        """
        return self.iteration_counter

    def set_iteration_counter(self, count: int) -> None:
        """_summary_

        Parameters
        ----------
        count : int
            Iteration count
        """
        self.iteration_counter = count

    def advance_iteration_counter(self) -> None:
        """
        Adds one to the iteration counter.
        """
        self.iteration_counter += 1

    def get_from_cache(self, observer: Observer, key: str) -> List[Any]:
        """

        Gets from the cache with the given key. The observed is passed along
        aswell in order to ensure that the observer is allowed to get with
        that key.

        Parameters
        ----------
        observer : class object
            An AGOX Observer object, e.g. an instance of a Sampler.
        key : str
            The key with which to get something from the cache.

        Returns
        -------
        list
            List of things stored with the given key.
        """
        # Makes sure the module has said it wants to get with this key.
        assert key in observer.get_values
        return deepcopy(self.cache.get(key))

    def add_to_cache(self, observer: Observer, key: str, data: List[Any], mode: Literal["w", "a"]) -> None:
        """
        Add data to the cache.

        Parameters
        ----------
        observer : class object
            An AGOX Observer object, e.g. an instance of a Sampler.
        key : str
            The key with which to get something from the cache.
        data : list
            List of data to store in the cache.
        mode : str
            Determines the mode in which the data is added to the cache:
            w: Will overwrite existing data with the same key.
            a: Will append to existing data (if there is existing data).
        """
        assert type(data) is list
        # Makes sure the module has said it wants to get with this key.
        assert key in observer.set_values
        assert mode in ["w", "a"]

        if key in self.cache.keys() and mode != "w":
            self.cache[key] += data
        else:
            self.cache[key] = data

    def clear(self) -> None:
        """
        Clears the current cachce. Called at the end of each iteration.
        """
        self.cache = {}
        self.execution_times = OrderedDict()

    def get_convergence_status(self) -> bool:
        """
        Returns the convergence status.

        Returns
        -------
        bool
            If True convergence has been reached and the main iteration-loop
            will halt.
        """
        return self.converged

    def set_convergence_status(self, state: bool) -> None:
        """
        Set the convergence status.

        Parameters
        ----------
        state : bool
            If True convergence has been reached and the main iteration-loop
            will halt.
        """
        self.converged = state

    def set_execution_time(self, name: str, time: float) -> None:
        """
        Set the execution time of a module.

        Parameters
        ----------
        name : str
            The name of the module.
        time : float
            The execution time of the module.
        """
        self.execution_times[name] = time



from abc import ABC, abstractmethod
from typing import Any, List, Literal, Tuple, Union

import numpy as np

from agox.analysis.search_data import SearchCollection, SearchData


class PropertyData(ABC):
    def __init__(self, data: Any, name: str) -> None:
        self.data = data
        self.name = name


class ArrayPropertyData(PropertyData):
    def __init__(self, data: np.ndarray, name: str, shape: Tuple[str], array_axis: Tuple[np.ndarray]) -> None:
        """
        Multidimensional array property data.

        E.g. for energy this is a 2D array with shape (Restarts, Iterations) and each entry is the energy of the configuration.

        Parameters
        ----------
        property_name : str
            Name of the property
        property_shape : Tuple[str]
            Shape of the property, e.g ('Restarts', 'Iterations') for a 2D array
        """
        super().__init__(data=data, name=name)
        self.shape = shape
        self.axis = array_axis


class ListPropertyData(PropertyData):
    def __init__(self, data: List[Any], name: str, shape: Tuple[str], list_axis: Tuple[Any]) -> None:
        """
        Parameters
        ----------
        property_name : str
            Name of the property
        """
        super().__init__(data=data, name=name)
        self.shape = shape
        self.axis = list_axis

    def __repr__(self) -> str:
        return f"{self.name}:\n{self.shape}"


class Property(ABC):

    def __init__(self, 
                time_axis: np.ndarray = None, 
                time_unit: Literal["s", "m", "h", "d", "y"] = "s", 
                cost_factor: int = 1) -> None:
        """
        Get the energy of the system as a np.array of shape [restarts, iterations].
        """
        if time_axis is None:
            time_axis = "indices"

        if time_axis not in ["indices", "iterations", "time"]:
            raise ValueError("time_axis must be one of ['indices', 'iterations', 'time']")

        self.time_axis = time_axis
        self.time_unit_symbol = time_unit
        
        if time_unit == "s":
            self.unit = 1
        elif time_unit == "m":
            self.unit = 60
        elif time_unit == "h":
            self.unit = 3600
        elif time_unit == "d":
            self.unit = 86400        
        elif time_unit == "y":
            self.unit = 31536000
        else:
            raise ValueError("time_unit must be one of ['s', 'm', 'h', 'd']")
        
        self.cost_factor = cost_factor

    def get_time_axis(self, search_data: SearchData) -> str:
        axis_name = self.time_axis.capitalize()

        if self.time_axis == "indices":
            indices = search_data.get_all("indices", fill=np.nan)
        elif self.time_axis == "iterations":
            indices = search_data.get_all("iterations", fill=np.nan)
        else:
            indices = search_data.get_all("times", fill=np.nan) / self.unit * self.cost_factor
            if self.cost_factor == 1:
                axis_name = axis_name + f" [{self.time_unit_symbol}]"
            else:
                axis_name = f"CPU cost [{self.time_unit_symbol}]"

        if self.time_axis in ["indices", "iterations"]:
            axis_name = f"{axis_name} [#]"

        return axis_name, indices

    @abstractmethod
    def compute(self, search_data: SearchData) -> PropertyData:
        pass

    def __call__(self, search_data: SearchData) -> PropertyData:
        return self.compute(search_data)

    def get_minimum(self, searches: Union[SearchCollection, List[SearchData]]) -> float:
        min_value = np.inf
        for search in searches:
            prop = self(search)
            if isinstance(prop, ArrayPropertyData):
                min_value = min(min_value, np.nanmin(prop.data))
            else:
                raise ValueError(f"Property {prop} is not an ArrayPropertyData")

        return min_value



import numpy as np
from ase.calculators.calculator import all_changes

from agox.acquisitors.ABC_acquisitor import AcquisitonCalculatorBaseClass, AcquisitorBaseClass


class LowerConfidenceBoundAcquisitor(AcquisitorBaseClass):
    """
    Lower Confidence Bound Acquisitor, using predictions and uncertainties from a model.

    The acquisition function is given by

    .. math::

        F(x) = E(x) - \kappa \sigma(x)


    Parameters
    -----------

    model: agox.models.ABC_model.ModelBaseClass
        Model that is used for the acquisition function.
    kappa: float
        Constant that is multiplied with the uncertainty.
    kwargs: dict
        Dictionary of keyword arguments passed to the AcquisitorBaseClass.
    """

    name = "LCBAcquisitor"

    def __init__(self, model, kappa=1, **kwargs):
        super().__init__(**kwargs)
        self.kappa = kappa
        self.model = model

    def calculate_acquisition_function(self, candidates):
        fitness = np.zeros(len(candidates))

        # Attach calculator and get model_energy
        for i, candidate in enumerate(candidates):
            E, sigma = self.model.predict_energy_and_uncertainty(candidate)
            fitness[i] = self.acquisition_function(E, sigma)

            # For printing:
            candidate.add_meta_information("model_energy", E)
            candidate.add_meta_information("uncertainty", sigma)

        return fitness

    def print_information(self, candidates, acquisition_values):

        format_float = lambda f: f"{f:.3f}"

        columns = ['Candidate', 'Energy', 'Uncertainty', 'Fitness', 'Generator']
        rows = []
        if self.model.ready_state:
            for i, candidate in enumerate(candidates):
                fitness = format_float(acquisition_values[i])
                Emodel = format_float(candidate.get_meta_information("model_energy"))
                sigma = format_float(candidate.get_meta_information("uncertainty"))
                generator = candidate.get_meta_information("generator")
                rows.append([str(i), Emodel, sigma, fitness, generator])
        
        self.writer.write_table(columns, rows, show_lines=False, show_edge=False)

    def get_acquisition_calculator(self):
        return LowerConfidenceBoundCalculator(self.model, self.acquisition_function, self.acquisition_force)

    def acquisition_function(self, E, sigma):
        return E - self.kappa * sigma

    def acquisition_force(self, E, F, sigma, sigma_force):
        return F - self.kappa * sigma_force

    def do_check(self, **kwargs):
        return self.model.ready_state


class LowerConfidenceBoundCalculator(AcquisitonCalculatorBaseClass):
    implemented_properties = ["energy", "forces"]

    def __init__(self, model, acquisition_function, acquisition_force, **kwargs):
        super().__init__(model, **kwargs)
        self.acquisition_function = acquisition_function
        self.acquisition_force = acquisition_force

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        super().calculate(atoms, properties, system_changes)

        derivatives = "forces" in properties
        model_data = self.model.converter(atoms, derivatives=derivatives)

        E = self.model.predict_energy(atoms, **model_data)
        sigma = self.model.predict_uncertainty(atoms, **model_data)
        self.results["energy"] = self.acquisition_function(E, sigma)

        if derivatives:
            F = self.model.predict_forces(atoms, **model_data)
            sigma_force = self.model.predict_uncertainty_forces(atoms, **model_data)

            self.results["forces"] = self.acquisition_force(E, F, sigma, sigma_force)



"""
An environment describes the search problem
- How many and which types of atoms will the search place?
- What is the computational cell?
- Is there a fixed template, e.g. a surface to be decorated by the search?
- Is the search constrained to a certain region of space?

These are settings that are not part of the search algorithm itself, but
rather describe the problem that the search is trying to solve.
"""

from agox.environments.environment import Environment

from .ABC_environment import EnvironmentBaseClass

__all__ = ["EnvironmentBaseClass", "Environment"]




import numpy as np
from ase.data import atomic_numbers
from matscipy.neighbours import neighbour_list as msp_nl
from scipy.spatial.distance import cdist

from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass


class ACSF(DescriptorBaseClass):
    name = "ACSF"
    descriptor_type = "local"

    def __init__(self, eta, rs, rc=5, xi=[], lamb=[], eta_ang=[], **kwargs):
        """
        Class for calculation of Behler Paranello features.

        Inputs:
           -- species: List of atomic numbers or atomic symbols
           -- eta: Exponent of radial functions
           -- rs: Offset of Gaussians.
           -- rc: Cutoff radius
           -- xi: Polynomial exponents of angular functions
           -- lamb (da): Signs for angular functions
           -- eta_ang: Exponential exponents of angular functions.
        """

        super().__init__(**kwargs)
        self.species = sorted([atomic_numbers[i] for i in self.environment.get_species()])

        self.eta = np.array(eta)
        self.rs = np.array(rs)
        self.xi = np.array(xi)
        self.lamb = np.array(lamb)
        self.eta_ang = np.array(eta_ang)
        self.rc = rc

        # Dimensions:
        if isinstance(eta, list):
            self.num_radial = len(eta)
        elif isinstance(rs, list):
            self.num_radial = len(rs)
        else:
            self.num_radial = 1

        self.num_angular = len(xi)

        self.dimension = (self.num_radial + self.num_angular) * len(self.species) + 1

    def create_features(self, atoms):
        """
        Calculates the Behler Paranello feature vectors for all atoms in a given structure.
        """
        # Matrix to save output:
        num_atoms = len(atoms)
        F = np.zeros((num_atoms, self.dimension))

        # Calculate distances:
        i1, i2, dists = msp_nl("ijd", atoms, self.rc)
        dists = dists.reshape(-1, 1)
        # Radial functions given by:
        for i in range(num_atoms):
            args = i1 == i
            relevant_dists = dists[args].reshape(-1, 1)
            for c, species in enumerate(self.species):
                filter = atoms.numbers[i2[args]] == species
                d = relevant_dists[filter]
                F[i, c * self.num_radial : (c + 1) * self.num_radial] = np.sum(
                    np.exp(-self.eta * (d - self.rs) ** 2) * self.cutoff(d), axis=0
                )

        F[:, -1] = atoms.numbers
        return F

        for i in range(num_atoms):
            neigh_mask = i1 == i
            for Rij, rij, j in zip(D[neigh_mask], dists[neigh_mask].ravel(), i2[neigh_mask]):
                for Rik, rik, k in zip(D[neigh_mask], dists[neigh_mask].ravel(), i2[neigh_mask]):
                    # print(i, j, k)
                    # if j != k and k != i: Equal conditions?
                    if j < k:
                        Rjk = Rik - Rij
                        rjk = np.sqrt(Rjk @ Rjk)
                        theta = (Rij @ Rik) / (np.sqrt(Rij @ Rij) * np.sqrt(Rik @ Rik))
                        F[i, self.num_radial : :] += (
                            (1 - self.lamb * theta) ** self.xi
                            * np.exp(-self.eta_ang * (rij**2 + rik**2 + rjk**2) / rc2)
                            * self.cutoff(rij)
                            * self.cutoff(rik)
                            * self.cutoff(rjk)
                        )
        F[:, self.num_radial : :] *= 2 ** (1 - self.xi)

        return F

    def calculate_specifics(self, atoms, index):
        """
        Arguments:
        -- Atoms: Atoms object
        -- Index: List/np.array of indicies of the atoms to calculate features for.
        """

        if type(index) == int or type(index) == np.int64:
            index = [index]

        F = np.zeros((len(index), self.dimension), dtype=float)

        num_atoms = len(atoms)
        for ii, idx in enumerate(index):
            allowed_atoms = [i for i in range(num_atoms) if i != idx]

            d = cdist(atoms.positions, atoms.positions[idx].reshape(1, -1))
            F[ii, 0 : self.num_radial] = np.sum(
                (np.exp(-self.eta * (d[allowed_atoms] - self.rs) ** 2 / self.rc**2) * self.cutoff(d[allowed_atoms])),
                axis=0,
            )

            for j in range(num_atoms):
                if j == idx:
                    continue
                Rij = atoms.positions[j, :] - atoms.positions[idx, :]
                rij = d[j]
                Fij = self.cutoff(rij)
                for k in range(j + 1, num_atoms):
                    if k == idx:
                        continue
                    Rjk = atoms.positions[k, :] - atoms.positions[j, :]
                    Rik = atoms.positions[k, :] - atoms.positions[idx, :]
                    theta = (Rij @ Rik) / (np.sqrt(Rij @ Rij) * np.sqrt(Rik @ Rik))
                    rjk = np.sqrt(Rjk @ Rjk)
                    rik = d[k]
                    Fik = self.cutoff(rik)
                    Fjk = self.cutoff(rjk)
                    F[ii, self.num_radial : :] += (
                        (1 - self.lamb * theta) ** self.xi
                        * np.exp(-self.eta_ang * (rij**2 + rik**2 + rjk**2) / self.rc**2)
                        * Fij
                        * Fjk
                        * Fik
                    )

            F[:, self.num_radial : :] *= 2 ** (1 - self.xi)
        return F

    def calculate_jacobian(self, atoms):
        """
        Calculate the Jacobian of the feature vector at the current positions
        """

        N = len(atoms)

        # Arrays for saving:
        J = np.zeros((self.dimension * N, 3 * N))

        # Calculate distances:
        rij = cdist(atoms.positions, atoms.positions)
        X = atoms.positions

        for i in range(N):
            for j in range(N):
                if rij[i, j] < self.rc:
                    for x in range(3):
                        if i != j:
                            J[i * self.dimension : (i + 1) * self.dimension, j * 3 + x] = (
                                np.exp(-self.eta * (rij[i, j] - self.rs) ** 2 / self.rc**2)
                                * (
                                    -2 * self.eta * (rij[i, j] - self.rs) / self.rc**2 * self.cutoff(rij[i, j])
                                    + self.cutoff_derivative(rij[i, j])
                                )
                                * (X[i, x] - X[j, x])
                                / rij[i, j]
                            )

                            J[i * self.dimension : (i + 1) * self.dimension, i * 3 + x] = J[
                                i * self.dimension : (i + 1) * self.dimension, j * 3 + x
                            ]

        return J

    def cutoff(self, r):
        return (r <= self.rc) * 0.5 * (1 + np.cos(np.pi * r / self.rc))

    def cutoff_derivative(self, r):
        return -0.5 * np.pi / self.rc * np.sin(np.pi * r / self.rc)

    @classmethod
    def from_species(cls, species, **kwargs):
        from ase import Atoms

        from agox.environments import Environment

        environment = Environment(
            template=Atoms(""), symbols="".join(species), use_box_constraint=False, print_report=False
        )
        return cls(environment=environment, **kwargs)

    def get_number_of_centers(self, atoms):
        return len(atoms)



from agox.analysis.property.descriptor_property import DescriptorProperty
from agox.analysis.property.energy import EnergyProperty
from agox.analysis.property.free_energy import FreeEnergyProperty
from agox.analysis.property.property import ArrayPropertyData, ListPropertyData, Property, PropertyData



from functools import partial
from typing import List

import numpy as np

from agox.candidates import StandardCandidate
from agox.models.ABC_model import ModelBaseClass
from agox.samplers.ABC_sampler import SamplerBaseClass
from agox.utils.thermodynamics import ThermodynamicsData, gibbs_free_energy


class GibbsFreeEnergySampler(SamplerBaseClass):
    name = "GFESampler"

    def __init__(
        self,
        model: ModelBaseClass,
        thermo_data: ThermodynamicsData,
        sample_size: int = 20,
        gets: dict = {"get_key": "evaluated_candidates"},
        order: int = 6,
        **kwargs,
    ):
        super().__init__(gets=gets, order=order, **kwargs)
        self.sample_size = sample_size
        self.model = model
        self.gfe = partial(gibbs_free_energy, thermo_data=thermo_data)

    def get_random_member(self) -> StandardCandidate:
        idx = np.random.randint(0, len(self.sample))
        return self.sample[idx].copy()

    def setup(self, all_candidates: List[StandardCandidate]) -> None:
        idx = np.argsort([self.gfe(c, self.model.predict_energy(c)) for c in all_candidates])
        self.sample = [all_candidates[i] for i in idx][: min(self.sample_size, len(idx))]



"""
Evaluators are responsible for evaluating the objective function of a candidate.
The main two flavours of evaluators are single-point and local optimization.

Single-point evaluators evaluate the objective function at a single point,
whereas local optimization evaluators perform a local optimization starting from
the candidate's coordinates.
"""
# ruff: noqa: I001, E402
from typing import TypeAlias
from .ABC_evaluator import EvaluatorBaseClass
Evaluator: TypeAlias = EvaluatorBaseClass

from .local_optimization import LocalOptimizationEvaluator
from .rattle import RattleEvaluator
from .single_point import SinglePointEvaluator

__all__ = [
    "EvaluatorBaseClass",
    "Evaluator",
    "SinglePointEvaluator", 
    "LocalOptimizationEvaluator", 
    "RattleEvaluator"
    ]



"""
Generators are are responsible for generating new configurations.
Typically there are two types of generators:

* Those that can generate a candidate by modifying one or more configurations.
* Those that can generate a candidate without any input configurations.

The first catagory of generators require a Sampler to sample the configurations
that it uses as a basis for generating a new candidate. The second category of
generators do not require a sampler.

The generators have to generate a candidate that follows the rules set by the
environment, that is

* The correct stoichoimetry
* Inside the confinement box.
"""
# ruff: noqa: I001, E402

from typing import TypeAlias
from agox.generators.ABC_generator import GeneratorBaseClass
Generator: TypeAlias = GeneratorBaseClass

from agox.generators.ce_generator import ComplementaryEnergyGenerator
from agox.generators.cog import CenterOfGeometryGenerator
from agox.generators.MD import MDgenerator
from agox.generators.permutation import PermutationGenerator
from agox.generators.random import RandomGenerator
from agox.generators.rattle import RattleGenerator
from agox.generators.replace import ReplaceGenerator
from agox.generators.reuse import ReuseGenerator
from agox.generators.sampling import SamplingGenerator
from agox.generators.steepest_descent import SteepestDescentGenerator
from agox.generators.symmetry import SymmetryGenerator
from agox.generators.symmetry_permutation import SymmetryPermutationGenerator
from agox.generators.symmetry_rattle import SymmetryRattleGenerator
from agox.generators.add_remove import AddRemoveGenerator

__all__ = ["GeneratorBaseClass", 
           "RandomGenerator", 
           "RattleGenerator", 
           "ReplaceGenerator",
           "ReuseGenerator", 
           "PermutationGenerator",
           "SteepestDescentGenerator", 
           "CenterOfGeometryGenerator", 
           "SamplingGenerator", 
           "ComplementaryEnergyGenerator", 
           "SymmetryGenerator", 
           "SymmetryRattleGenerator", 
           "SymmetryPermutationGenerator", 
           "MDGenerator",
           "AddRemoveGenerator"]




import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional

import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator, all_changes

from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.databases import Database
from agox.main import State
from agox.observer import Observer, ObserverHandler
from agox.utils.filters import FilterBaseClass


class ModelBaseClass(Calculator, ObserverHandler, Observer, ABC):
    """Model Base Class implementation

    Parameters
    ----------
    database : AGOX Database obj
        If used for on-the-fly training on a database, this should be set
    iteration_start_training : int
        When model is attached as observer it starts training after this number of
        iterations.
    update_period : int
        When model is attached as observer it updates every update_period
        iterations.
    record : set
        Training record.

    """

    def __init__(
        self,
        database: Optional[Database] = None,
        filter: FilterBaseClass = None, # noqa
        order: float = 0,
        iteration_start_training: int = 0,
        update_period: int = 1,
        surname: str = "",
        gets: dict[str, str] | None = None,
        sets: dict[str, str] | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        database : AGOX Database obj
            If used for on-the-fly training on a database, this should be set
        filter : FilterBaseClass
            Filter object to use for filtering data
        order : float
            Order of observer
        iteration_start_training : int
            When model is attached as observer it starts training after this number of
            iterations.
        update_period : int
            When model is attached as observer it updates every update_period
            iterations.
        surname : str
            Surname of observer
        gets : dict
            Gets dictionary
        sets : dict
            Sets dictionary    
        """
        Observer.__init__(self, order=order, surname=surname, gets=gets, sets=sets, **kwargs)
        ObserverHandler.__init__(self, handler_identifier="model", dispatch_method=self.training_observer)
        Calculator.__init__(self)

        self.filter = filter
        self._save_attributes = ["_ready_state"]
        self.iteration_start_training = iteration_start_training
        self.update_period = update_period

        self.validation_data = []
        self._ready_state = False
        self._record = set()
        self.update = False

        self.add_observer_method(
            self.training_observer,
            gets=self.gets[0],
            sets=self.sets[0],
            order=self.order[0],
            handler_identifier="database",
        )

        if database is not None:
            self.attach_to_database(database)

    @property
    @abstractmethod
    def name(self) -> None:  # pragma: no cover
        """str: Name of model. Must be implemented in child class."""
        ...

    @property
    @abstractmethod
    def implemented_properties(self) -> None:  # pragma: no cover
        """:obj: `list` of :obj: `str`: Implemented properties.
        Available properties are: 'energy', 'forces', 'uncertainty'

        Must be implemented in child class.
        """
        ...

    @abstractmethod
    def predict_energy(self, atoms: Atoms, **kwargs) -> None:  # pragma: no cover
        """Method for energy prediction.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float
            The energy prediction

        Must be implemented in child class.
        """
        pass

    @abstractmethod
    def train(self, training_data: list[Atoms], **kwargs) -> None:  # pragma: no cover
        """Method for model training.

        Note
        ----------
        Always include **kwargs when implementing this function.
        If your model is not trainable just write a method that does nothing

        Parameters
        ----------
        atoms : :obj: `list` of :obj: `ASE Atoms`
            List of ASE atoms objects or AGOX candidate objects to use as training data.
            All atoms must have a calculator with energy and other nesseary properties set, such that
            it can be accessed by .get_* methods on the atoms.


        Must be implemented in child class.

        """
        pass

    @property
    def ready_state(self) -> None:
        """bool: True if model has been trained otherwise False."""
        return self._ready_state

    @ready_state.setter
    def ready_state(self, state: bool) -> None:
        self._ready_state = bool(state)

    def add_save_attributes(self, attribute: str | list[str]) -> None:
        """Add attribute to save list.

        Parameters
        ----------
        attribute : str or list of str
            Name of attribute to add to save list.

        """
        if isinstance(attribute, str):
            self._save_attributes.append(attribute)
        else:
            self._save_attributes += attribute

    def remove_save_attributes(self, attribute: str | list[str]) -> None:
        """Remove attribute from save list.

        Parameters
        ----------
        attribute : str or list of str
            Name of attribute to remove from save list.

        """
        if isinstance(attribute, str):
            self._save_attributes.remove(attribute)
        else:
            for a in attribute:
                self._save_attributes.remove(a)

    def reset_save_attributes(self) -> None:
        """Reset save list."""
        self._save_attributes = []

    @Observer.observer_method
    def training_observer(self, database: Database, state: State) -> None:
        """Observer method for use with on-the-fly training based data in an AGOX database.

        Note
        ----------
        This implementation simply calls the train_model method with all data in the database

        Parameters
        ----------
        atoms : AGOX Database object
            The database to keep the model trained against

        Returns
        ----------
        None

        """
        iteration = self.get_iteration_counter()

        if iteration < self.iteration_start_training:
            return
        if iteration % self.update_period != 0 and iteration != self.iteration_start_training:
            return

        data = database.get_all_candidates()
        self.train(data)
        self.dispatch_to_observers(model=self, state=state)

    def add_validation_data(self, data: list[Atoms]) -> None:
        """Add validation data to model.

        Parameters
        ----------
        data : :obj: `list` of :obj: `ASE Atoms`
            List of ASE atoms objects or AGOX candidate objects to use as validation data.
            All atoms must have a calculator with energy and other nesseary properties set, such that
            it can be accessed by .get_* methods on the atoms.

        """
        if isinstance(data, list):
            self.validation_data += data
        else:
            self.validation_data.append(data)

    def predict_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        """Method for forces prediction.

        The default numerical central difference force calculation method is used, but
        this can be overwritten with an analytical calculation of the force.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array
            The force prediction with shape (N,3), where N is len(atoms)

        """
        return self.predict_forces_central(atoms, **kwargs)

    def predict_forces_central(self, atoms: Atoms, acquisition_function: Optional[Callable] = None, d: float=0.001, **kwargs) -> np.ndarray:
        """Numerical cenral difference forces prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.
        acquisition_function : Acquisition function or None
            Function that takes evaluate acquisition function based on
            energy and uncertainty prediction. Used for relaxation in acquisition
            funtion if force uncertainties are not available.

        Returns
        ----------
        np.array
            The force prediction with shape (N,3), where N is len(atoms)

        """
        if acquisition_function is None:
            energy = lambda a: self.predict_energy(a) # noqa
        else:
            energy = lambda a: acquisition_function(*self.predict_energy_and_uncertainty(a)) # noqa

        e0 = energy(atoms)  # self.predict_energy(atoms)
        energies = []

        for a in range(len(atoms)):
            for i in range(3):
                new_pos = atoms.get_positions()  # Try forward energy
                new_pos[a, i] += d
                atoms.set_positions(new_pos)
                if atoms.positions[a, i] != new_pos[a, i]:  # Check for constraints
                    energies.append(e0)
                else:
                    energies.append(energy(atoms))
                    atoms.positions[a, i] -= d

                new_pos = atoms.get_positions()  # Try backwards energy
                new_pos[a, i] -= d
                atoms.set_positions(new_pos)
                if atoms.positions[a, i] != new_pos[a, i]:
                    energies.append(e0)
                else:
                    energies.append(energy(atoms))
                    atoms.positions[a, i] += d

        penergies = np.array(energies[0::2])  # forward energies
        menergies = np.array(energies[1::2])  # backward energies

        forces = ((menergies - penergies) / (2 * d)).reshape(len(atoms), 3)
        return forces

    def predict_uncertainty(self, atoms: Atoms, **kwargs) -> float:
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float
            The energy uncertainty prediction

        """
        return 0

    def predict_uncertainty_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array
            The force uncertainty prediction with shape (N,3) with N=len(atoms)

        """
        return np.zeros((len(atoms), 3))

    def predict_energy_and_uncertainty(self, atoms: Atoms, **kwargs) -> tuple[float, float]:
        """Method for energy and energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        float, float
            The energy and energy uncertainty prediction

        """
        return self.predict_energy(atoms, **kwargs), self.predict_uncertainty(atoms, **kwargs)

    def predict_forces_and_uncertainty(self, atoms: Atoms, **kwargs) -> tuple[np.ndarray, np.ndarray]:
        """Method for energy and energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        np.array, np.array
            Forces and forces uncertainty. Both with shape (N, 3) with N=len(atoms).

        """
        return self.predict_forces(atoms, **kwargs), self.predict_forces_uncertainty(atoms, **kwargs)

    def converter(self, atoms: Atoms, **kwargs) -> dict:
        """Converts an ASE atoms object to a format that can be used by the model

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for which to predict the energy.

        Returns
        ----------
        object
            The converted object

        """
        return {}

    def calculate(self, atoms: Atoms = None, properties: list[str]=["energy"], system_changes: list[str] = all_changes) -> None:
        """ASE Calculator calculate method

        Parameters
        ----------
        atoms : ASE Atoms obj or AGOX Candidate object
            The atoms object for to predict properties of.
        properties : :obj: `list` of :obj: `str`
            List of properties to calculate for the atoms
        system_changes : ASE system_changes
            Which ASE system_changes to check for before calculation

        Returns
        ----------
        None
        """
        Calculator.calculate(self, atoms, properties, system_changes)

        energy = self.predict_energy(self.atoms)
        self.results["energy"] = energy

        if "forces" in properties:
            forces = self.predict_forces(self.atoms)
            self.results["forces"] = forces

    def validate(self, **kwargs) -> dict:
        """Method for validating the model.

        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to the validation method.

        Returns
        ----------
        dict
            Dictionary with validation results

        """
        if len(self.validation_data) == 0:
            return None

        e_true = []
        e_pred = []
        for d in self.validation_data:
            e_true.append(d.get_potential_energy())
            e_pred.append(self.predict_energy(d))

        e_true = np.array(e_true)
        e_pred = np.array(e_pred)

        return {
            "Energy MAE [eV]": np.mean(np.abs(e_true - e_pred)),
            "Energy RMSE [eV]": np.sqrt(np.mean((e_true - e_pred) ** 2)),
            "Max absolute energy error [eV]": np.max(np.abs(e_true - e_pred)),
            "Max relative energy error [%]": np.max((e_true - e_pred) / e_true) * 100,
            "Min relative energy error [%]": np.min((e_true - e_pred) / e_true) * 100,
        }

    def _training_record(self, data: list[Atoms]) -> None:
        """
        Record the training data.

        Parameters
        ----------
        data : list
            List of Atoms objects.

        """
        if not all([isinstance(d, CandidateBaseClass) for d in data]):
            return

        for d in data:
            self._record.add(d.cache_key)

        self.update = True

    def _get_new_data(self, data: list[Atoms]) -> tuple[list, list]:
        """
        Get the new data.

        Parameters
        ----------
        data : list
            List of Atoms objects.

        Returns
        -------
        list
            List of new Atoms objects.

        list
            List of old Atoms objects.

        """
        if not all([isinstance(d, CandidateBaseClass) for d in data]):
            return data, []

        new_data = []
        old_data = []
        for d in data:
            if d.cache_key in self._record:
                old_data.append(d)
            else:
                new_data.append(d)
        return new_data, old_data

    def print_model_info(self, validation: dict[str, Any] = None, **kwargs) -> None:
        """Prints model information

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to pass to the model

        Returns
        ----------
        None
        """
        model_info = self.model_info(**kwargs)
        if validation is not None:
            model_info.append("------ Validation Info ------")
            model_info.append("Validation data size: {}".format(len(self.validation_data)))
            for key, val in validation.items():
                model_info.append("{}: {:.3}".format(key, val))

        for s in model_info:
            self.writer(s)

    def model_info(self, **kwargs) -> list[str]:
        """Returns model information

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to pass to the model

        Returns
        ----------
        list of str
            The model information
        """
        return ["No model information available."]

    def save(self, path: str = "model.h5") -> None:
        """
        Save the model as a pickled object.

        Parameters
        ----------
        path : str, optional
            Path to save the model to. The default is 'model.h5'.
        """
        import h5py

        with h5py.File(path, "w") as f:
            for key in self._save_attributes:
                obj = self
                for k in key.split("."):
                    data = getattr(obj, k)
                    obj = data
                if data is not None:
                    f.create_dataset(key, data=data)

        self.writer("Saving model to {}".format(path))

    def load(self, path: str) -> None:
        """
        Load a pickle

        Parameters
        ----------
        path : str
            Path to a saved model.

        Returns
        -------
        model-object
            The loaded model object.
        """
        assert os.path.exists(path), "Path does not exist"
        import h5py

        self.writer("Loading model from {}".format(path))
        with h5py.File(path, "r") as f:
            for key in self._save_attributes:
                obj = self
                for k in key.split(".")[:-1]:
                    obj = getattr(obj, k)

                k = key.split(".")[-1]

                if key in f:
                    value = f[key][()]
                    setattr(obj, k, value)

        self.writer("Model loaded")

    def attach_to_database(self, database: Database) -> None:
        from agox.databases.ABC_database import DatabaseBaseClass

        assert isinstance(database, DatabaseBaseClass)
        print(f"{self.name}: Attaching to database: {database}")
        self.attach(database)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms
from ase.build import fcc100
from ase.calculators.emt import EMT

from agox import AGOX
from agox.acquisitors import GibbsFreeEnergyAquisitor
from agox.collectors import ParallelCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import AddRemoveGenerator, RandomGenerator, RattleGenerator
from agox.models.descriptors.soap import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import GFEKMeansSampler
from agox.utils.sparsifiers import CUR
from agox.utils.thermodynamics import ThermodynamicsData

seed = 42
database_index = 0

##############################################################################
# System & general settings:
##############################################################################

# Define a template surface using ASE functions.
template = fcc100("Cu", size=(3, 3, 2), vacuum=5.0)
template.pbc = [True, True, False]
template.positions[:, 2] -= template.positions[:, 2].min()

# Confinement cell matches the template cell in x & y but is smaller in z.
confinement_cell = template.cell.copy()
confinement_cell[2, 2] = 4.0

# Confinement corner is at the origin of the template cell, except in z where
# it is shifted up.
z0 = template.positions[:, 2].max()
confinement_corner = np.array([0, 0, z0])  # Confinement corner is at cell origin

# The environment now specifies a range of how many atoms of each species can be
# present in the system.
environment = Environment(
    template=template,
    symbols_range={"Cu": [5, 6, 7], "O": [3, 4]},
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=7)

##############################################################################
# Calculator
##############################################################################

calc = EMT()

################################################################################
# Reference energies.
################################################################################

# References energies should be correctly calculated, e.g. from molecular oxygen
# and energy differences between Cu-slabs of different thickness.
references = {"Cu": 0, "O": -1.0}
chemical_potentials = {"Cu": 0, "O": -1.0}  # Choice of chemical potential.
thermo_data = ThermodynamicsData(references, chemical_potentials)

##############################################################################
# Search Settings:
##############################################################################

descriptor = SOAP(
    environment=environment,
    r_cut=6,
    nmax=3,
    lmax=2,
    sigma=1,
    weight=True,
    periodic=True,
)
kernel = C(1, (1, 100)) * RBF(20, (10, 30))

single_atom_energies = np.zeros(100)
for symbol in references.keys():
    number = Atoms(symbol).numbers[0]
    single_atom_energies[number] = references[symbol]

model = SparseGPR(
    descriptor=descriptor,
    kernel=kernel,
    database=database,
    sparsifier=CUR(1000),  
    sigma=0.01,
    prior=Repulsive(),
    train_uncertainty=True,
    single_atom_energies=single_atom_energies,
)

sample_size = 10
sampler = GFEKMeansSampler(
    thermo_data=thermo_data,
    model=model,
    descriptor=descriptor,
    database=database,
    sample_size=sample_size,
)
rattle_generator = RattleGenerator(**environment.get_confinement())
random_generator = RandomGenerator(**environment.get_confinement(), contiguous=False)
addrm_generator = AddRemoveGenerator(**environment.get_confinement())

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator, addrm_generator]
num_candidates = {0: [10, 0, 0], 5: [2, 4, 4]}

# Acquisitor that uses the Gibbs free energy to select candidates.
acquisitor = GibbsFreeEnergyAquisitor(
    thermo_data=thermo_data,
    model=model,
    kappa=1,
    order=3,
)
# CPU-count is set here for Ray - leave it out to use as many cores as are available.
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)
# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5},
    start_relax=8,
    order=2,
)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=True,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 0},
    order=6,
)
##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(collector, acquisitor, relaxer, database, evaluator, seed=seed)

agox.run(N_iterations=10)



import os
from pathlib import Path

from ase.calculators.calculator import Calculator, all_changes, all_properties


class SubprocessGPAW(Calculator):
    """
    GPAW calculator that uses a subprocess to run the calculation.

    Parameters
    -------------
    ncores (int): The number of cores to use.
        If None, the number of cores is determined by the number of cores available.

    **kwargs:
        Additional keyword arguments to pass to the GPAW for calculation settings.
        Eg.
            kwargs = {
                'mode': {'name': 'pw', 'ecut': cutoff},
                'xc': 'PBE',
        Is a planewave PBE calculation.
        The same dict can be passed directly to the normal GPAW calculator, to exactly
        reproduce the calculation settings.
    """

    implemented_properties = ["energy", "forces", "stress"]

    def __init__(self, ncores=None, log_directory="gpaw_logs/", **kwargs):
        super().__init__()
        self.kwargs = kwargs

        if ncores is None:
            ncores = len(os.sched_getaffinity(0))
        self.ncores = ncores
        self.count = 0

        self.log_directory = Path(log_directory)
        if not self.log_directory.exists():
            self.log_directory.mkdir(parents=True)

    def calculate(self, atoms=None, properties=all_properties, system_changes=all_changes):
        from ase.calculators.subprocesscalculator import gpaw_process  # Because of ASE versions.

        super().calculate(atoms, properties, system_changes)

        # Update the log file name.
        kwargs = self.kwargs.copy()
        if kwargs.get("txt", "-") != "-":
            name = Path(kwargs["txt"]).stem + f"_{self.count}.txt"
            kwargs["txt"] = str(self.log_directory / name)

        # Run the calculation.
        with gpaw_process(ncores=self.ncores, **kwargs) as gpaw:
            gpaw._run_calculation(self.atoms.copy(), properties, system_changes)
            results = gpaw.protocol.recv()

        self.results.update(results)
        self.count += 1



from cgitb import handler
import matplotlib

matplotlib.use("Agg")

import numpy as np
from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator
from ase import Atoms

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# General settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template, symbols="Au5Ni", confinement_cell=confinement_cell, confinement_corner=confinement_corner
)

# Database
db_path = "db1.db"
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=False,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 2},
    order=2,
)

##############################################################################
# Let get the show running!
##############################################################################

from agox.observer import Observer

class DatabaseAndMainObserver(Observer):
    name = "DatabaseAndMainObserver"

    def __init__(self, order=[0, 0]):
        Observer.__init__(self, order=order)
        self.add_observer_method(
            self.database_observer_method, order=self.order[0], sets={}, gets={}, handler_identifier="database"
        )
        self.add_observer_method(
            self.main_observer_method, order=self.order[0], sets={}, gets={}, handler_identifier="AGOX"
        )

    @Observer.observer_method
    def database_observer_method(self, database, state):
        size_of_database = len(database)
        self.writer(f"Database size: {size_of_database}")
        self.writer(f"I am run from the database, I see the database {database} and {state}.")

    @Observer.observer_method
    def main_observer_method(self, state):
        self.writer(f"I am run from main, I see only the state: {state}")


database_and_main_observer = DatabaseAndMainObserver()
database_and_main_observer.attach(database)

agox = AGOX(random_generator, database, evaluator, database_and_main_observer)

agox.run(N_iterations=1)

# database.observer_reports()
# database.print_observers()



import numpy as np


def atomic_numerical_derivative(func, atoms, delta=1e-3):
    N_atoms = len(atoms)

    f_output_shape = func(atoms).shape
    F = np.zeros((N_atoms, 3, *f_output_shape))

    stencil = [-2, -1, 0, 1, 2]

    for a in range(N_atoms):
        for d in range(3):
            f = {}
            for s in stencil:
                atoms_tmp = atoms.copy()
                atoms_tmp.positions[a, d] += delta * s
                f[s] = func(atoms_tmp)

            F[a, d] = (1 * f[-2] - 8 * f[-1] + 0 * f[0] + 8 * f[1] - 1 * f[2]) / (12 * delta)

    return -F



from typing import Union, Tuple, List
from pathlib import Path

import pytest

from agox.utils.thermodynamics import ThermodynamicsData


@pytest.mark.parametrize('element', ['Cu', 29])
def test_thermodata_get_copper(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    assert thermo_data.get_chemical_potential(element) == -1.0

@pytest.mark.parametrize('element', ['O', 8])
def test_thermodata_get_oxygen(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    assert thermo_data.get_chemical_potential(element) == -2.0

@pytest.mark.parametrize('element', ['O', 8])
def test_thermodata_set_oxygen(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.set_chemical_potential(element, -3.0)
    assert thermo_data.get_chemical_potential('O') == -3.0
    thermo_data.set_chemical_potential(element, -2.0)
    assert thermo_data.get_chemical_potential('O') == -2.0

@pytest.mark.parametrize('element', ['Cu', 29])
def test_thermodata_set_copper(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.set_chemical_potential(element, -3.0)
    assert thermo_data.get_chemical_potential('Cu') == -3.0
    thermo_data.set_chemical_potential(element, -1.0)
    assert thermo_data.get_chemical_potential('Cu') == -1.0

@pytest.mark.parametrize('element', ['Al', 13])
def test_thermodata_get_aluminium(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    with pytest.raises(KeyError):
        thermo_data.get_chemical_potential(element)

@pytest.mark.parametrize('element', ['Cu', 29])
def test_thermodata_get_ref_copper(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.get_reference(element) == -1.0

@pytest.mark.parametrize('element', ['O', 8])
def test_thermodata_get_ref_oxygen(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.get_reference(element) == -2.0

@pytest.mark.parametrize('element', ['O', 8])
def test_thermodata_set_ref_oxygen(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.set_reference(element, -3.0)
    assert thermo_data.get_reference('O') == -3.0
    thermo_data.set_reference(element, -2.0)
    assert thermo_data.get_reference('O') == -2.0

@pytest.mark.parametrize('element', ['Cu', 29])
def test_thermodata_set_ref_copper(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    thermo_data.set_reference(element, -3.0)
    assert thermo_data.get_reference('Cu') == -3.0
    thermo_data.set_reference(element, -1.0)
    assert thermo_data.get_reference('Cu') == -1.0

@pytest.mark.parametrize('element', ['Al', 13])
def test_thermodata_get_ref_aluminium(element: Union[str, int], thermo_data: ThermodynamicsData) -> None:
    with pytest.raises(KeyError):
        thermo_data.get_reference(element)

@pytest.mark.parametrize('element_result', ([29, 'Cu'], [8, 'O'], ['Cu', 'Cu'], ['O', 'O'], [13, 'Al'], ['Al', 'Al']))
def test_thermodata_convert_element(element_result: Tuple[List[Union[str, int]]], thermo_data: ThermodynamicsData) -> None:
    element = element_result[0]
    assert thermo_data.convert_number_to_element(element) == element_result[1]

def test_thermodata_template_energy(thermo_data: ThermodynamicsData) -> None:
    assert thermo_data.template_energy == -10.0

def test_thermodata_save(thermo_data: ThermodynamicsData, tmpdir: Path) -> None:
    thermo_data.save(tmpdir / 'thermo_data.json')
    thermo_data2 = ThermodynamicsData.load(tmpdir / 'thermo_data.json')

    assert thermo_data.references == thermo_data2.references
    assert thermo_data.chemical_potentials == thermo_data2.chemical_potentials
    assert thermo_data.template_energy == thermo_data2.template_energy

def test_thermodata_save_raise(thermo_data: ThermodynamicsData, tmpdir: Path) -> None:
    thermo_data.save(tmpdir / 'thermo_data.json')

    with pytest.raises(FileExistsError):
        thermo_data.save(tmpdir / 'thermo_data.json')

def test_thermodata_save_overwrite(thermo_data: ThermodynamicsData, tmpdir: Path) -> None:
    thermo_data.save(tmpdir / 'thermo_data.json')
    thermo_data.save(tmpdir / 'thermo_data.json', overwrite=True)







from typing import Tuple

import numpy as np

from agox.analysis.criterion.base_criterion import BaseCriterion
from agox.analysis.property.property import PropertyData


class ThresholdCriterion(BaseCriterion):
    def __init__(self, threshold: float):
        """
        Find the time at which the property first crosses below a threshold.

        Parameters
        ----------
        threshold : float

        """
        self.threshold = threshold

    def compute(self, property_data: PropertyData) -> Tuple[np.array, np.array]:
        """
        Parameters
        ----------
        property_arr : np.array
            The property array to compute the metric on, shape (n_restarts, n_observations).
        """
        property_arr = property_data.data
        property_time = property_data.axis[1]

        # Rolling minimum of the property array
        min_energies = np.minimum.accumulate(property_arr, axis=1)

        # Check if the minimum energy is below the threshold
        success = min_energies <= self.threshold

        # Find the time of the first success
        events = np.max(success, axis=1).astype(int)
        times_indices = np.argmax(success, axis=1)

        times = np.array([property_data.axis[1][i, time_index] for i, time_index in enumerate(times_indices)])
        times[events == 0] = np.nanmax(property_time)

        return times, events



from abc import ABC, abstractmethod

from agox.observer import Observer
from agox.writer import Writer


class AttractorMethodBaseClass(ABC, Observer):
    """
    Baseclass for attractors methods

    Parameters
    -----------
    order: int
        When to update the structures by checking the database
    """

    dynamic_attributes = ["structures"]

    def __init__(self, order=10):
        Observer.__init__(self, order=order)
        self.add_observer_method(self.update, order=self.order[0], sets={}, gets={})

        self.structures = []

    @Observer.observer_method
    def update(self, database, state):
        self.structures = database.get_all_candidates()

    @abstractmethod
    def get_attractors(self, structure):
        pass



import numpy as np

from agox.acquisitors.ABC_acquisitor import AcquisitorBaseClass


class MetaInformationAcquisitor(AcquisitorBaseClass):
    name = "MetaInformationAcquisitor"

    def __init__(self, meta_key, mode="min", **kwargs):
        super().__init__(**kwargs)
        self.meta_key = meta_key
        self.mode = mode

        assert mode in ["max", "min"]  # Either maximize or minimize according to the meta key.

    def calculate_acquisition_function(self, candidates):
        sign = 1 if self.mode == "min" else -1
        fitness = np.zeros(len(candidates))

        for i, candidate in enumerate(candidates):
            fitness[i] = sign * candidate.get_meta_information(self.meta_key)

        return fitness



from typing import List, Optional

import numpy as np
import torch
from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.calculators.mixing import SumCalculator as SumCalculatorASE

from agox import Module


class BlockCalculator(Calculator):
    implemented_properties = ["energy", "forces"]

    def __init__(self, blocks: List[Atoms], power: float = 4.0, factor: float = 1.0) -> None:
        super().__init__()
        self.blocks = blocks
        self.process_blocks(blocks)

        self.power = power
        self.factor = factor

    def calculate(self, atoms: Atoms, properties: Optional[List] = None, system_changes: Optional[List] = None) -> None:
        super().calculate(atoms, properties, system_changes)

        # Extract information about the blocks
        blocks_used = atoms.meta_information["blocks_used"]
        block_indices = atoms.meta_information["block_indices"]

        # Get the block indices
        edge_index = torch.tensor(self.get_edge_indices(block_indices, blocks_used), dtype=torch.long)
        positions = torch.tensor(atoms.get_positions(), requires_grad=True)

        # Calculate the distances
        r = positions[edge_index[:, 0]] - positions[edge_index[:, 1]]
        r_norm = torch.norm(r, dim=1)
        r_blocks = self.get_block_distances(blocks_used)

        # Calculate the energy
        r_diff = r_norm - r_blocks
        block_bond_energy = (self.factor * r_diff) ** self.power 
        energy = torch.sum(block_bond_energy)

        # Calculate the forces
        forces = -torch.autograd.grad(energy, positions)[0]

        self.results["energy"] = energy.detach().numpy().item()
        self.results["forces"] = forces.detach().numpy()

    def get_edge_indices(self, block_indices: np.ndarray, blocks_used: np.ndarray) -> np.ndarray:
        stacked_block_indices = []

        for atoms_idx, block_idx in zip(block_indices, blocks_used):
            stacked_block_indices.append(self.edge_indices[block_idx] + min(atoms_idx))

        z = np.concatenate(stacked_block_indices, axis=0).astype(np.int32)
        return z

    def get_initial_index_array(self, block_indices: np.ndarray) -> np.ndarray:
        stacked_block_indices = []

        for block_index in block_indices:
            i, j = np.meshgrid(block_index, block_index)
            tri_index = np.triu_indices(len(block_index), k=1)
            ij = np.stack([i[tri_index].flatten(), j[tri_index].flatten()], axis=1)
            stacked_block_indices.append(ij)

        z = np.concatenate(stacked_block_indices, axis=0).astype(np.int32)
        return z

    def process_blocks(self, blocks: List[Atoms]) -> None:
        """
        Could add a cutoff in this function so that not all inter-block distances are 
        calculated. This could allow blocks to flex. 
        """
        block_distances = []
        edge_indices = []

        for block in blocks:
            edge_index = self.get_initial_index_array([np.arange(len(block))])
            positions = torch.tensor(block.get_positions())
            r = positions[edge_index[:, 0]] - positions[edge_index[:, 1]]
            r_norm = torch.norm(r, dim=1)
            block_distances.append(r_norm)
            edge_indices.append(edge_index)

        self.block_distances = block_distances
        self.edge_indices = edge_indices

    def get_block_distances(self, blocks_used: np.ndarray) -> torch.Tensor:
        block_distances = torch.concatenate([self.block_distances[i] for i in blocks_used])
        return block_distances


class SumCalculator(SumCalculatorASE, Module):

    def __init__(self, calc1: Calculator, calc2: Calculator) -> None:
        SumCalculatorASE.__init__(self, calcs=[calc1, calc2])
        Module.__init__(self)
        self.calc1 = calc1
        self.calc2 = calc2
        self.ready_state = True










import numpy as np
from agox.utils.ray_utils import RayPoolUser


class ParallelSum(RayPoolUser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def sum(self, X):
        # Make the jobs:
        N_jobs = X.shape[0]
        modules = [[]] * N_jobs
        args = [[x] for x in X]
        kwargs = [{} for _ in range(N_jobs)]

        # Run the jobs:
        results = self.pool_map(np.sum, modules, args, kwargs)

        return np.array(results)


X = np.random.rand(100, 100)

parallel_sum = ParallelSum(cpu_count=4)
results = parallel_sum.sum(X)

assert (results == np.sum(X, axis=1)).all()



import warnings

import numpy as np
from scipy.stats import CensoredData, ecdf


class DiscreteDistribution:
    def __init__(self, times: np.ndarray, events: np.ndarray):
        self.succcesful = times[np.where(events == 1)]
        self.censored = times[np.where(events == 0)]

        self.rv = CensoredData(self.succcesful, right=self.censored)
        self.ecdf = ecdf(self.rv)

        self.times = times
        self.events = events

    def cdf(self) -> np.ndarray:
        cdf = self.ecdf.cdf.probabilities
        cdf = np.concatenate([[0], cdf])
        return cdf

    def quantiles(self) -> np.ndarray:
        quantiles = self.ecdf.cdf.quantiles
        quantiles = np.concatenate([[0], quantiles])
        return quantiles

    def lower(self, alpha: float = 0.95) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            lower = self.ecdf.cdf.confidence_interval(alpha).low.probabilities

        lower = np.concatenate([[0], lower])
        return lower

    def upper(self, alpha: float = 0.95) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            upper = self.ecdf.cdf.confidence_interval(alpha).high.probabilities
        upper = np.concatenate([[0], upper])
        return upper

    def pmf(self) -> np.ndarray:
        pmf = np.zeros_like(np.unique(self.times))

        for i, t in enumerate(np.unique(self.times)):
            pmf[i] = np.sum(self.events[self.times == t])

        pmf = pmf / len(self.events)

        return pmf



import numpy as np
import pytest

from agox.candidates import StandardCandidate
from agox.utils.thermodynamics import ThermodynamicsData
from agox.utils.thermodynamics.gibbs import gibbs_free_energy


def test_gibbs_candidate(candidate: StandardCandidate, thermo_data: ThermodynamicsData, expected_energy: float) -> None:
    assert gibbs_free_energy(candidate, thermo_data=thermo_data) == expected_energy

def test_gibbs_numbers(numbers: np.ndarray, total_energy: int, thermo_data: ThermodynamicsData, expected_energy: float) -> None:
    assert gibbs_free_energy(numbers=numbers, total_energy=total_energy, thermo_data=thermo_data) == expected_energy

def test_gibbs_candidate_dicts(candidate: StandardCandidate, references: dict, chemical_potentials: dict, template_energy: float, expected_energy: float) -> None:
    assert gibbs_free_energy(candidate, references=references, chemical_potentials=chemical_potentials, template_energy=template_energy) == expected_energy

def test_gibbs_raise(candidate: StandardCandidate) -> None:

    with pytest.raises(ValueError):
        gibbs_free_energy(candidate)



from typing import List, Union

import numpy as np
from matplotlib.axes import Axes

from agox.analysis.criterion import BaseCriterion
from agox.analysis.property import Property
from agox.analysis.search_data import SearchCollection, SearchData


class SuccessPlotter:
    def __init__(
        self, searches: Union[SearchCollection, List[SearchData]], search_property: Property, criterion: BaseCriterion
    ) -> None:
        self.searches = searches
        self.search_property = search_property
        self.criterion = criterion

    def plot(self, ax: Axes) -> None:
        for search in self.searches:
            self.plot_search(ax, search)

        ax.set_xlim(0, None)
        ax.set_ylim(0, 1)
        ax.legend()

    def plot_search(self, ax: Axes, search: SearchData) -> None:
        # Gather stuff:
        prop = self.search_property(search)
        distribution = self.criterion(prop)

        # Plot:
        (l1,) = ax.step(distribution.quantiles(), distribution.cdf(), label=f"{search.get_label()}", where="post")
        ax.fill_between(distribution.quantiles(), distribution.lower(), distribution.upper(), step="post", alpha=0.3)

        # Plot where the searches have gotten to
        points = np.nanmax(prop.axis[1], axis=1)
        y = distribution.ecdf.cdf.evaluate(points)
        ax.plot(points, y, "x", color=l1.get_color())

        ax.set_xlabel(f"{prop.shape[1]}")
        ax.set_ylabel("Success")



"""
Models are used to predict the output of a given input. Generally this corresponds
to predicting the energy and forces of a given atomic configuration.

This module also contains the descriptors used to represent the atomic configurations.
"""
# ruff: noqa: I001, E402
from typing import TypeAlias
from .ABC_model import ModelBaseClass
Model: TypeAlias = ModelBaseClass

from agox.models.GPR import GPR, SparseGPR, SparseGPREnsemble
from agox.models.composition_model import CompositionModel
from agox.models.ase_model import CalculatorModel

__all__ = ['ModelBaseClass', 'Model', 'GPR', 'SparseGPR', 'SparseGPREnsemble', 'CompositionModel', 'CalculatorModel']




from typing import List, Literal, Optional

import numpy as np

from agox.candidates import Candidate
from agox.main.state import State
from agox.models import Model, ModelBaseClass
from agox.observer import Observer
from agox.samplers import SamplerBaseClass
from agox.samplers.replica_exchange import RateTracker, ReplicaExchangeSample


class ReplicaExchangeSampler(SamplerBaseClass):
    name = "RepliceExchangeSampler"

    """
    Replica exchange sampler.

    Parameters
    ----------
    sample_size : int
        Number of members in the sample.
    swap : str
        Method for swapping members. Options are 'up' or 'down'.
    temperatures: List[float]
        List of temperatures to use.
    t_max : float
        Maximum temperature.
    t_min : float
        Minimum temperature.
    swap_interval : int
        Number of iterations between swap attempts.
    swap: Literal["up", "down"]
        Method for swapping members. Options are 'up' or 'down'.
    always_accept : bool
        If True, always accept the new candidate.
    flush_prob : float
        Probability of flushing a member of the sample - to replace it by a new candidate.
        Happens after the model has been updated.
    gets : dict
        Dictionary with the key to get the candidates.
    sets : dict
        Dictionary with the key to set the candidates.
    """

    def __init__(
        self,
        model: Model = None,
        t_min: float = None,
        t_max: float = None,
        sample_size: int = 5,
        swap: Literal["up", "down"] = "up",
        temperatures: Optional[np.ndarray] = None,
        swap_interval: int = 10,
        always_accept: bool = False,
        flush_prob: float = 0.0,
        gets: dict = {"get_key": "candidates", "get_evaluated": "evaluated_candidates"},
        sets: dict = {"set_key": "candidates", "sample_key": "sample"},
        flush_minima: bool = False,
        force_criterion: float = 0.05,
        **kwargs,
    ) -> None:
        super().__init__(gets=gets, sets=sets, **kwargs)

        self._set_temperatures(t_min, t_max, sample_size, temperatures)
        self.sample_size = sample_size
        self.swap_interval = swap_interval
        self.always_accept = always_accept
        self.flush_prob = flush_prob
        self.flush_minima = flush_minima
        self.force_criterion = force_criterion

        self.sample = ReplicaExchangeSample(sample_size)
        self.tracker = RateTracker(sample_size)
        self.reset_observer()  # We remove observers added by the base-class.
        self._add_observers()

        if swap == "up":
            self.swap = self.swap_up
        elif swap == "down":
            self.swap = self.swap_down
        else:
            raise ValueError(f"Swap method '{swap}' not recognized. Use 'up' or 'down'.")

        if model is not None:
            self.attach_to_model(model)

    ############################################################################
    # Observer methods
    ############################################################################

    @Observer.observer_method
    def update_sampler_energies(self, model: Model, state: State) -> None:
        self.writer.debug(
            "Sampler Update", f"{'E0':^8s}", f"{'Edft':^8s}", f"{'Edft-E0':^8s}", f"{'E1':^8s}", f"{'E1-E0':^8s}"
        )

        evaluated_candidates = state.get_from_cache(self, self.get_evaluated)
        dft_uuids = {}
        for cand in evaluated_candidates:
            dft_uuids[cand.get_meta_information("uuid")] = [
                cand.get_meta_information("evaluator_energy"),
                cand.get_meta_information("evaluator_max_force"),
            ]

        iteration = self.get_iteration_counter()
        for i, sample_member in enumerate(self.sample):
            e_dft = np.nan
            if not sample_member.empty():
                uuid = sample_member.candidate.get_meta_information("uuid")
                if uuid in dft_uuids:
                    if iteration in dft_uuids[uuid][0]:
                        e_dft = dft_uuids[uuid][0][iteration]
                e_old = sample_member.get_energy()

                energy = model.predict_energy(sample_member.candidate)  # noqa: N806
                sample_member.energy = energy

            self.writer.debug(
                f"Sample {i} == {sample_member.candidate.get_meta_information('walker_index')}:",
                f"{e_old:8.3f}",
                f"{e_dft:8.3f}",
                f"{e_dft - e_old:8.3f}",
                f"{energy:8.3f}",
                f"{energy - e_old:8.3f}",
            )

        if self.flush_minima:
            for i in range(len(self.sample)):
                uuid = self.sample[i].candidate.get_meta_information("uuid")
                max_force = None
                if uuid in dft_uuids:
                    if iteration in dft_uuids[uuid][1]:
                        max_force = dft_uuids[uuid][1][iteration]
                print(max_force)
                if max_force is not None:
                    if max_force < self.force_criterion:
                        self.sample.update(i, None, np.inf)

        if self.flush_prob > 0.0:
            for i in range(len(self.sample)):
                if np.random.rand() < self.flush_prob:
                    self.sample.update(i, None, np.inf)

    @Observer.observer_method
    def setup_sampler(self, state: State) -> None:
        if self.do_check():
            evaluated_candidates = state.get_from_cache(self, self.get_key)
            evaluated_candidates = list(filter(None, evaluated_candidates))
            if len(evaluated_candidates) > 0:
                self.setup(evaluated_candidates)

        # Overwrite the candidates on the cache as the sampler may have updated meta information.
        state.add_to_cache(self, self.set_key, evaluated_candidates, mode="w")

        sample = [sample_member.candidate for sample_member in self.sample]
        state.add_to_cache(self, self.sample_key, sample, mode="w")

    ############################################################################
    # Setup methods
    ############################################################################

    def setup(self, evaluated_candidates: List[Candidate]) -> None:
        self.update_sample(evaluated_candidates)

        iteration = self.get_iteration_counter()
        if self.decide_to_swap(iteration):
            self.swap()

        if iteration % 1 == 0:
            self.report_statistics()

    def update_sample(self, evaluated_candidates: List[Candidate]) -> None:
        for candidate in evaluated_candidates:
            walker_index = candidate.get_meta_information("walker_index")
            accepted = False
            if self.sample[walker_index].empty():  # Happens either when flushed or when starting.
                accepted = True
            else:  # Regular Metropolis check when not empty.
                parent = self.sample[walker_index]
                temperature = self.temperatures[walker_index]
                accepted = self.metropolis_check(candidate, parent, temperature)

            if accepted:
                energy = candidate.get_potential_energy()
                self.sample.update(walker_index, candidate, energy)

            self.tracker.update_acceptance(walker_index, int(accepted))
            candidate.add_meta_information("accepted", accepted)

    def metropolis_check(self, candidate: Candidate, parent: Candidate, temperature: float) -> bool:
        if self.always_accept:
            return True
        E_candidate = candidate.get_potential_energy()  # noqa
        E_parent = parent.get_energy()  # noqa - Parent is a SampleMember

        if E_candidate < E_parent:
            accepted = True
            P = 1
        else:
            P = np.exp(-(E_candidate - E_parent) / temperature)  # noqa
            accepted = P > np.random.rand()

        return accepted

    ############################################################################
    # Swap methods
    ############################################################################

    def swap_check(self, i: int, j: int) -> bool:
        """
        Check if we should swap the i-th and j-th member of the sample.

        Parameters
        ----------
        i: int
            Index of the first member.
        j: int
            Index of the second member.

        Returns
        -------
        bool
            True if the swap is accepted, False otherwise.
        """
        E_i = self.sample[i].get_energy()  # noqa: N806
        E_j = self.sample[j].get_energy()  # noqa: N806
        beta_i = 1 / self.temperatures[i]
        beta_j = 1 / self.temperatures[j]

        P = np.min([1, np.exp((beta_i - beta_j) * (E_i - E_j))])  # noqa: N806

        return P > np.random.rand()

    def swap_up(self) -> None:
        self.writer("Swapping in 'up' mode")

        # Run over the sample starting from the bottom:
        for i in range(len(self.sample) - 1):
            j = i + 1
            swap_bool = self.swap_check(i, j)

            self.tracker.update_swap_up(i, int(swap_bool))
            self.tracker.update_swap_down(j, int(swap_bool))

            if swap_bool:
                self.sample.swap(i, j)

    def swap_down(self) -> None:
        self.writer("Swapping in 'down' mode")

        # Run over the sample starting from the highest temperature:
        for i in range(len(self.sample) - 1, 0, -1):
            j = i - 1
            swap_bool = self.swap_check(i, j)

            self.tracker.update_swap_up(j, int(swap_bool))
            self.tracker.update_swap_down(i, int(swap_bool))

            if swap_bool:
                self.sample.swap(i, j)

    def decide_to_swap(self, iteration_counter: int) -> bool:
        return iteration_counter % self.swap_interval == 0

    ############################################################################
    # Convenience methods
    ############################################################################

    def report_statistics(self) -> None:
        def float_format(value: float) -> str:
            return f"{value:.2f}"

        columns = ["Member", "Temperature", "Energy", "Accept", "Up", "Down"]
        rows = []
        energies = []
        acceptances = []
        for i in range(self.sample_size):
            swap_up = float_format(self.tracker.get_swap_up_rate(i))
            swap_down = float_format(self.tracker.get_swap_down_rate(i))
            a = self.tracker.get_acceptance_rate(i, start=-10)
            acceptance = float_format(a)
            e = self.sample[i].get_energy()
            energy = float_format(e)
            rows.append([str(i), float_format(self.temperatures[i]), energy, acceptance, swap_up, swap_down])
            acceptances.append(a or 0)
            energies.append(e or 0)

        self.writer.write_table(table_columns=columns, table_rows=rows, expand=True)
        self.writer.debug("Walker energies:   ", " ".join([f"{energy:8.3f}" for energy in energies]))
        self.writer.debug(
            "Walker strengths:  ", " ".join([f"{amplitude:8.3f}" for amplitude in self.tracker.rattle_amplitudes])
        )
        self.writer.debug("Walker acceptances:", " ".join([f"{acceptance:8.3f}" for acceptance in acceptances]))

    def get_walker(self, walker_index: int) -> Candidate:
        walker = self.sample[walker_index]
        if walker.empty():
            return None
        else:
            return walker.candidate.copy()

    def attach_to_model(self, model: Model) -> None:
        assert isinstance(model, ModelBaseClass)
        print(f"{self.name}: Attaching to model: {model}")
        self.attach(model)

    def _set_temperatures(self, t_min: float, t_max: float, sample_size: int, temperatures: np.ndarray) -> None:
        # Temperature logic:
        # Check that they are not all not none:
        if t_max is not None and t_min is not None and temperatures is not None:
            raise ValueError("Either t_max and t_min or temperatures must be provided, not both.")
        elif t_max is not None and t_min is not None:  # If both are provided, generate the temperatures.
            self.temperatures = np.geomspace(t_min, t_max, sample_size)
        elif temperatures is not None:
            if len(temperatures) != sample_size:
                raise ValueError(
                    f"Length of temperatures ({len(temperatures)}) does not match sample size ({sample_size})."
                )
            else:
                self.temperatures = temperatures
        else:
            raise ValueError("Either t_max and t_min or temperatures must be provided.")

    def _add_observers(self) -> None:
        self.add_observer_method(
            self.update_sampler_energies,
            gets=self.gets[0],
            sets=self.sets[0],
            order=self.order[0],
            handler_identifier="model",
        )

        self.add_observer_method(
            self.setup_sampler,
            gets=self.gets[0],
            sets=self.sets[0],
            order=self.order[0],
            handler_identifier="AGOX",
        )



# Copyright 2021-2024, Mads-Peter V. Christiansen, Bjørk Hammer, Nikolaj Rønne.
# This file is part of AGOX.
# AGOX is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# AGOX is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should
# have received a copy of the GNU General Public License along with AGOX. If not, see <https://www.gnu.org/licenses/>.

import atexit  # noqa
from datetime import datetime
from typing import Iterable

import numpy as np

from agox.candidates.standard import CandidateBaseClass, StandardCandidate
from agox.observer import FinalizationHandler, Observer, ObserverHandler
from agox.writer import Writer, get_icon, set_writer_verbosity
from agox.main.state import State


class AGOX(ObserverHandler, FinalizationHandler):
    """
    AGO-X
    Atomistic Global Optimization X
    """

    name = "AGOX"

    def __init__(
        self,
        *args,
        seed: int = None,
        candidate_instanstiator: CandidateBaseClass = None,
        attach: bool = True,
        report: bool = True,
        verbosity: int | None = None,
    ) -> None:
        """
        Parameters
        ----------
        *args : Observer
            Observer objects to attach to AGOX.
        seed : int, optional
            Seed for the random number generator.
        candidate_instanstiator : CandidateBaseClass, optional
            Candidate instanstiator to use, default is StandardCandidate.
        attach : bool, optional
            If True the observers are attached to AGOX.
        report : bool, optional
            If True a report is printed.
        """
        ObserverHandler.__init__(self, handler_identifier="AGOX", dispatch_method=self.run)
        FinalizationHandler.__init__(self)

        if verbosity is not None:
            set_writer_verbosity(verbosity)

        self._observers = list(args)
        self.candidate_instanstiator = candidate_instanstiator or StandardCandidate
        self.writer = Writer()

        if seed is not None:
            self.set_seed(seed)

        # This attaches all Observers
        if attach:
            self.update()

        if report:
            self.report()

    def report(self) -> None:
        icon = get_icon()
        print(icon)
        self.print_observers(hide_log=True)
        self.observer_reports(hide_log=True)

    def update(self) -> None:
        """
        Calls 'attach' on all Observer-objects in 'self.elements' and updates
        the candidate_instanstiator on all Observer-objects in 'self.elements'
        """
        for observer in self._observers:
            if hasattr(observer, "attach"):
                observer.attach(self)
            if hasattr(observer, "set_candidate_instanstiator"):
                observer.set_candidate_instanstiator(self.candidate_instanstiator)

    def set_seed(self, seed: int) -> None:
        if seed is not None:
            np.random.seed(seed)
            self.writer("Numpy random seed: {}".format(seed))

    def print_iteration_header(self, state: "State") -> None:
        print("\n")
        self.writer.write_header("Iteration: {}".format(state.get_iteration_counter()))
        dt = datetime.now()
        self.writer("Time: {}".format(dt.strftime("%H:%M:%S")))
        self.writer("Date: {}".format(dt.strftime("%d/%m/%Y")))

    def run(self, N_iterations: int, verbose=True, hide_log=True) -> None:  # noqa
        """
        Function called by runscripts that starts the actual optimization procedure.

        This function is controlled by modules attaching themselves as observers to this module.
        The order system ensures that modules with lower order are executed first, but no gurantee within each order,
        so if two modules attach themselves both with order 0 then their individual execution order is not guranteed.
        However, an observer with order 0 will always be executed before an observer with order 1.

        The default ordering system is:
        order = 0: Execution order

        All modules that intend to attach themselves as observers MUST take the order as an argument (with a default value(s)),
        so that if a different order is wanted that can be controlled from runscripts. Do NOT change order default values!
        """

        # Main-loop calling the relevant observers.
        state = State()

        # Force print if something crashes the run.
        atexit.register(self.on_failure)

        self.writer.write_header("AGOX run started")

        while state.get_iteration_counter() <= N_iterations and not state.get_convergence_status():
            self.print_iteration_header(state)
            self.dispatch_to_observers(state=state)
            self.print_timings(state)

            state.clear()
            state.advance_iteration_counter()
            self.writer.write_header("Iteration finished")

        # Some things may want to perform some operation only at the end of the run.
        for method in self.get_finalization_methods():
            method()

        # Remove the force print if everything went well.
        atexit.unregister(self.on_failure)

        self.writer.write_header("AGOX run finished")

    def add_observer(self, observer: Observer) -> None:
        """
        Add an observer to the AGOX object.

        Parameters
        ----------
        observer : Observer
            An observer to add to AGOX.
        """

        if not isinstance(observer, Observer):
            raise TypeError(f"Observer must be an instance of Observer: {observer.__name__}")

        self._observers.append(observer)

    def add_observers(self, *observers: Iterable[Observer], update: bool = True, report: bool = True) -> None:
        """
        Add a list of modules to the AGOX object.

        Parameters
        ----------
        modules : list
            A list of modules to be added to the AGOX object.
        """

        for observer in observers:
            self.add_observer(observer)

        if update:
            self.update()

        if report:
            self.report()

    def on_failure(self) -> None:
        """
        This method is called if the the main loop crashes.
        """
        print("RUN FAIL: AGOX run failed.")

    def print_timings(self, state: "State") -> None:
        from rich.table import Table
        from rich.tree import Tree

        self.writer.write_header("Timings")
        total_time = 0
        tree_list = []
        for name, time in state.execution_times.items():
            handler = name.split(".")[0]
            observer = ".".join(name.split(".")[1:2])
            method = ".".join(name.split(".")[2:])        
            tree_list.append([handler, observer, method, time])

            if handler == "AGOX":
                total_time += time

        row_table = Table.grid(expand=True)
        row_table.add_column()
        row_table.add_column(justify="right")
        row_table.add_row("Total time", f"{total_time:05.2f}s [   %  ]")
        tree = Tree(row_table, highlight=True)
        branches = {}
        for index in range(len(tree_list)):
            handler, observer, method, time = tree_list[index]
            percentage = time / total_time * 100
            text1 = f"{observer}.{method}"
            text2 = f"{time:05.2f}s [{percentage:05.2f}%]"

            row_table = Table.grid(expand=True)
            row_table.add_column()
            row_table.add_column(justify="right")
            row_table.add_row(text1, text2)

            if observer not in branches and handler not in branches:
                branches[observer] = tree.add(row_table)
            if handler in branches:
                branches[handler].add(row_table)

        self.writer(tree)






        



import numpy as np

from agox.generators.complementary_energy.attractor_methods.ABC_attractors import AttractorMethodBaseClass


class AttractorAnotherStructure(AttractorMethodBaseClass):
    """
    Method for picking attractors from another structure.

    Parameters
    -----------
    descriptor: Local descriptor
        Used to calculate local features
    possible_number_of_attractors: int or list with two ints
        If int, then that exact amount of attractors will be created. If a list,
        the number of attractors will be a random number between the two given numbers
    attractors_from_template: Bool
        Whether attractors can be picked as feature vectors of template atoms
    predefined_attractors: List of attractors
        If not None, the specified attractors will be given to the CE generator
    order: int
        When to update the structures by checking the database
    """

    name = "AttractorAnotherStructure"

    def __init__(
        self,
        descriptor,
        possible_number_of_attractors=[3, 6],
        attractors_from_template=True,
        predefined_attractors=None,
        order=5,
    ):
        super().__init__(order=order)
        self.descriptor = descriptor
        self.possible_number_of_attractors = possible_number_of_attractors
        self.attractors_from_template = attractors_from_template
        self.predefined_attractors = predefined_attractors

    def get_attractors(self, structure):
        if self.predefined_attractors is not None:
            return self.predefined_attractors

        template = structure.get_template()
        n_template = len(template)

        if isinstance(self.possible_number_of_attractors, int):
            number_of_attractors = self.possible_number_of_attractors
        else:
            number_of_attractors = np.random.randint(
                self.possible_number_of_attractors[0], self.possible_number_of_attractors[1]
            )

        database_structure_index = np.random.randint(len(self.structures))

        structure_from_database = self.structures[database_structure_index]

        features = self.descriptor.get_features(structure_from_database)
        if not self.attractors_from_template:
            features = features[n_template:]

        attractor_indices = np.random.choice(range(len(features)), size=number_of_attractors, replace=False)

        attractors = features[attractor_indices]

        return attractors



import ray

ray.init(num_cpus=4)


@ray.remote
class Actor:
    def __init__(self):
        self.x = 0

    def inc(self):
        self.x += 1
        return self.x

    def get_value(self):
        return self.x


actors = [Actor.remote() for i in range(4)]

values = ray.get([actor.get_value.remote() for actor in actors])

print(values)

actors[0].inc.remote()

values = ray.get([actor.get_value.remote() for actor in actors])

print(values)



import os
from time import sleep

import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import read, write

from agox.observer import Observer
from agox.samplers.metropolis import MetropolisSampler


def safe_write(filename, atoms):
    write(filename=filename + ".lock", images=atoms, format="traj")
    os.rename(filename + ".lock", filename)


class ConcurrentTemperingSampler(MetropolisSampler):
    """
    Parallel tempering sampler for concurrent searches with multiple worker jobs. Written to work with ConcurrentDatabase, no guarantee that it works in any other setting.

    Parameters
    ----------
    temperatures : list
        List of temperatures for each worker.
    database : ConcurrentDatabase
        The database that the sampler will attach to.
    swap_frequency : int
        The frequency at which the workers will swap candidates.
    gets : list
        List of dictionaries with the gets that the sampler will observe.
    sets : list
        List of dictionaries with the sets that the sampler will observe.
    swap_order : int
        Order of the observer.
    """

    name = "ParallelTemperingSampler"

    def __init__(
        self,
        temperatures=[],
        database=None,
        swap_frequency=10,
        gets=[{"get_key": "evaluated_candidates"}, {}],
        sets=[{}, {}],
        swap_order=2,
        verbose: bool = False,
        **kwargs,
    ):
        super().__init__(gets=gets, sets=sets, **kwargs)
        self.swap_frequency = swap_frequency
        self.temperatures = temperatures
        self.temperature = temperatures[database.worker_number]
        self.swap_order = swap_order
        self.swap_func = self.metropolis_hastings_swap
        self.most_recently_accepted = {worker_number: None for worker_number in range(database.total_workers)}
        self.most_recently_accepted_iteration = {worker_number: 0 for worker_number in range(database.total_workers)}
        self.verbose = verbose

        self.add_observer_method(
            self.swap_candidates,
            sets=self.sets[1],
            gets=self.gets[1],
            order=self.swap_order,
            handler_identifier="database",
        )

        self.attach_to_database(database)

    @Observer.observer_method
    def swap_candidates(self, database, state):
        """
        Assumes that the database is synced.
        """

        if self.decide_to_swap(database):
            for candidate in database.candidates:
                worker_number = candidate.get_meta_information("worker_number")
                iteration = candidate.get_meta_information("iteration")
                accepted = candidate.get_meta_information("accepted")

                if accepted:
                    if iteration > self.most_recently_accepted_iteration[worker_number]:
                        self.most_recently_accepted[worker_number] = candidate
                        self.most_recently_accepted_iteration[worker_number] = candidate.get_meta_information(
                            "iteration"
                        )

            if self.verbose:
                energies = [
                    self.most_recently_accepted[c].get_potential_energy()
                    if self.most_recently_accepted[c] is not None
                    else 0
                    for c in self.most_recently_accepted
                ]
                if database.worker_number == 0:
                    text = (
                        "PARALLEL TEMPERING:" + f"{self.temperature:8.3f}, " + ",".join([f"{e:8.3f}" for e in energies])
                    )
                else:
                    text = "PARALLEL TEMPERING (not main worker):"
                self.writer(text)

            self.swap_func(database)

    def metropolis_hastings_swap(self, database):
        worker_number = database.worker_number
        total_workers = database.total_workers
        iteration = self.get_iteration_counter()

        # I abuse the filenames a bit here:
        filename = database.filename[:-3] + "_swap_iteration_{}_worker_{}.traj"

        if worker_number == 0:
            # This one does the calculation and 'broadcasts' to the others over disk.

            # Starting from the bottom:
            for i in range(total_workers - 1):
                C_i = self.most_recently_accepted[i]
                C_j = self.most_recently_accepted[i + 1]

                E_i = C_i.get_potential_energy()
                E_j = C_j.get_potential_energy()

                beta_i = 1 / self.temperatures[i]
                beta_j = 1 / self.temperatures[i + 1]

                P = np.min([1, np.exp((beta_i - beta_j) * (E_i - E_j))])

                r = np.random.rand()

                if r < P:
                    self.writer("Swapped {} with {}".format(i, i + 1))
                    self.most_recently_accepted[i], self.most_recently_accepted[i + 1] = (
                        self.most_recently_accepted[i + 1],
                        self.most_recently_accepted[i],
                    )
                    self.most_recently_accepted_iteration[i], self.most_recently_accepted_iteration[i + 1] = (
                        iteration,
                        iteration,
                    )

                else:
                    self.writer("Did not swap {} for {}".format(i, i + 1))

            # Write the candidates:
            for wn in range(1, total_workers):
                atoms = self.most_recently_accepted[wn]
                atoms.info = atoms.meta_information
                safe_write(filename.format(iteration, wn), atoms)
            self.chosen_candidate = self.most_recently_accepted[worker_number]

        else:
            while not os.path.exists(filename.format(iteration, worker_number)):
                sleep(1)

            chosen_atoms = read(filename.format(iteration, worker_number))

            self.sample[0] = self.convert_to_candidate_object(chosen_atoms, self.sample[0].template)
            self.sample[0].meta_information = chosen_atoms.info
            scp = SinglePointCalculator(
                self.sample[0], energy=chosen_atoms.get_potential_energy(), forces=chosen_atoms.get_forces()
            )
            self.sample[0].set_calculator(scp)

            os.remove(filename.format(iteration, worker_number))

    def convert_to_candidate_object(self, atoms_type_object, template):
        candidate = self.candidate_instanstiator(
            template=template,
            positions=atoms_type_object.positions,
            numbers=atoms_type_object.numbers,
            cell=atoms_type_object.cell,
            info=atoms_type_object.info,
        )
        return candidate

    def decide_to_swap(self, database):
        return (self.get_iteration_counter() % self.swap_frequency == 0) * (database.total_workers > 1)



import rich_click as click


def convert_database_to_traj(database_path):
    import os

    from agox.databases import Database

    """
    Convert database to list of ASE Atoms objects. 

    Currently assumes that the database can be read using the standard database 
    class (agox/modules/databases/database.py).

    Parameters
    -----------
    database_path: str
        Path to database on disk. 

    Returns
    --------
    list
        List of ASE Atoms objects. 
    """

    if os.path.exists(database_path):
        database = Database(database_path, initialize=False)
    else:
        print(f"Trajectory not found: {database_path}")
        return []

    trajectory = database.restore_to_trajectory()

    return trajectory


def find_file_types(paths):
    """
    Find file extension of given paths.

    Parameters
    -----------
    paths: list of str
        Paths to check

    Returns
    --------
    list:
        List of strs indicating of the file types (extensions) of the given paths.
    """
    file_types = []
    for path in paths:
        file_types.append(path.split(".")[-1])
    return file_types


@click.command(name="convert")
@click.argument("files", nargs=-1)
@click.option("--name", "-n", default=None, type=str, help="Name of the output trajectory file.")
def cli_convert(files, name=None):
    """
    Convert database files to trajectory files.
    """
    import numpy as np
    from ase.io import write

    # Determine the file extension & ensure all paths have the same extension.
    file_types = find_file_types(files)
    assert np.array([file_type == file_types[0] for file_type in file_types]).all()
    file_type = file_types[0]

    # If database files - convert to trajectory.
    trajectory = []
    if file_type == "db":
        for path in files:
            print(f"Converting: {path}")
            trajectory += convert_database_to_traj(path)

        # Naming:
        if name is None:
            name = "converted_db.traj"

        # Save:
        print(f"Saving trajectory file {name}")
        write(name, trajectory)


if __name__ == "__main__":
    cli_convert()



import traceback
from uuid import uuid4

import numpy as np
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.optimize.bfgs import BFGS
from ase.optimize.optimize import Optimizer

from agox.candidates import Candidate

from .ABC_evaluator import EvaluatorBaseClass


class LocalOptimizationEvaluator(EvaluatorBaseClass):

    name = "LocalOptimizationEvaluator"

    def __init__(
        self,
        calculator: Calculator,
        optimizer: Optimizer = BFGS,
        optimizer_run_kwargs: dict | None = None,
        optimizer_kwargs: dict | None = None,
        fix_template: bool = True,
        constraints: list | None = [],
        store_trajectory: bool = True,
        **kwargs,
    ) -> None:
        """
        Evaluator that performs a local optimization of a candidate using a ASE BFGS optimizer.

        Parameters
        -----------
        calculator: `ase.calculators.calculator.Calculator`
            The calculator to use for the evaluation.
        optimizer: `ase.optimize.optimize.Optimizer`, optional
            The optimizer to use for the local optimization. Default is BFGS.
        optimizer_run_kwargs: dict, optional
            The keyword arguments to pass to the optimizer.run() method. Defaults to 
            `{"fmax": 0.25, "steps": 200}`.
        optimizer_kwargs: dict, optional
            The keyword arguments to pass to the optimizer constructor.
            Defaults to `{"logfile": None}`.
        fix_template: bool, optional
            Whether to fix the template atoms during the optimization. Default is True.
        constraints: list, optional
            List of constraints to apply during the optimization.
        store_trajectory: bool, optional
            Whether to store the trajectory of the optimization. Default is True.
            If False only the last step is stored.
        """
        super().__init__(**kwargs)
        self.calculator = calculator
        self.store_trajectory = store_trajectory

        # Optimizer stuff:
        self.optimizer = optimizer
        self.optimizer_kwargs = optimizer_kwargs if optimizer_kwargs is not None else {"logfile": None}
        self.optimizer_run_kwargs = optimizer_run_kwargs if optimizer_run_kwargs is not None else {"fmax": 0.25, "steps": 200}

        # Constraints:
        self.constraints = constraints

        if self.name == "LocalOptimizationEvaluator":
            if self.optimizer_run_kwargs["steps"] == 0:
                print(
                    f"{self.name}: Running with steps = 0 is equivalent to SinglePointEvaluator, consider using that instead to simplify your runscript."
                )
                self.store_trajectory = False

    def evaluate_candidate(self, candidate: Candidate) -> bool:
        candidate.calc = self.calculator
        try:
            if self.optimizer_run_kwargs["steps"] > 0:
                self.apply_constraints(candidate)
                optimizer = self.optimizer(candidate, **self.optimizer_kwargs)

                # The observer here stores all the steps if 'store_trajectory' is True.
                optimizer.attach(self._observer, interval=1, candidate=candidate, steps=optimizer.get_number_of_steps)

                optimizer.run(**self.optimizer_run_kwargs)
                candidate.add_meta_information("relax_index", optimizer.get_number_of_steps())

                steps = optimizer.get_number_of_steps()
                max_steps = self.optimizer_run_kwargs["steps"]
                E = candidate.get_potential_energy(apply_constraint=False) # noqa: N806
                self.writer(f"Optimization: {self.E_initial:5.3f} ---> {E:5.3f} ({steps}/{max_steps} steps)")

                # If 'store_trajectory' is False we manually store the last step.
                if not self.store_trajectory:
                    self._check_and_append_candidate(candidate)
            else:
                F = candidate.get_forces(apply_constraint=False) # noqa: N806
                E = candidate.get_potential_energy(apply_constraint=False) # noqa: N806
                self._check_and_append_candidate(candidate)

        except Exception as e:
            self.writer("Energy calculation failed with exception: {}".format(e))
            traceback.print_exc()
            return False

        E = candidate.get_potential_energy(apply_constraint=False) # noqa: N806
        F = candidate.get_forces(apply_constraint=False) # noqa: N806
        calc = SinglePointCalculator(candidate, energy=E, forces=F)
        candidate.calc = calc

        return True

    def _observer(self, candidate: Candidate, steps: int) -> None:
        self.check_callback(candidate)  # check before candidate gets copied and assigned a SinglePointCalculator

        E = candidate.get_potential_energy(apply_constraint=False) # noqa: N806
        if steps() == 0:
            self.E_initial = E
        F = candidate.get_forces(apply_constraint=False) # noqa: N806

        traj_candidate = candidate.copy()
        calc = SinglePointCalculator(traj_candidate, energy=E, forces=F)
        traj_candidate.set_calculator(calc)
        traj_candidate.add_meta_information("relax_index", steps())
        traj_candidate.add_meta_information("final", False)
        self._add_time_stamp(traj_candidate)

        if steps() > 0:
            traj_candidate.add_meta_information("uuid", str(uuid4()))

        if self.store_trajectory:
            self.evaluated_candidates.append(traj_candidate)

    def apply_constraints(self, candidate: Candidate) -> None:
        constraints = self.constraints
        candidate.set_constraint(constraints)    

        for constraint in candidate.constraints:
            if hasattr(constraint, 'reset'):
                constraint.reset() 

    def get_template_constraint(self, candidate: Candidate) -> FixAtoms:
        return FixAtoms(indices=np.arange(len(candidate.template)))



from agox.postprocessors.ABC_postprocess import PostprocessBaseClass


class WrapperPostprocess(PostprocessBaseClass):
    """
    Wraps a candidate object such that all atoms are in the  first unit cell.

    Parameters
    ----------
    None
    """

    name = "WrapperTheRapper"

    def postprocess(self, candidate):
        candidate.wrap()
        return candidate



from typing import Callable
from uuid import uuid4


class FinalizationHandler:
    """
    Just stores information about functions to be called when finalizaing a run.
    """

    def __init__(self) -> None:
        self.finalization_methods = {}
        self.names = {}

    def attach_finalization(self, name: str, method: Callable) -> None:
        """
        Attaches finalization method.

        Parameters
        -----------
        name: str
            Human readable name of the attached method.
        method: method
            A method or function to attach.
        """
        key = uuid4()
        self.finalization_methods[key] = method
        self.names[key] = name

    def get_finalization_methods(self) -> list:
        """
        Returns
        --------
        List
            List of finalization methods.
        """
        return self.finalization_methods.values()

    def print_finalization(self, include_observer: bool = False, verbose: int = 0) -> None:
        """
        include_observer flag might be useful to debug if something is not working as expected.
        """

        names = [self.names[key] for key in self.finalization_methods.keys()]

        base_string = "{}: order = {} - name = {} - method - {}"
        if include_observer:
            base_string += " - method: {}"

        print("=" * 24 + " Finalization " + "=" * 24)
        for name in names:
            print("Name: {}".format(name))
        print("=" * len("=" * 25 + " Observers " + "=" * 25))
